// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include "FlowUtils.h"
#include "MatrixOperations.h"

using namespace cv;
using namespace std;


const Vec2i kPOSITION_INVALID = Vec2i(-1, -1);
const float kMOVEMENT_UNKNOWN = -FLT_MAX;
const Vec2f kFLOW_UNKNOWN = Vec2f(kMOVEMENT_UNKNOWN,kMOVEMENT_UNKNOWN);


// Transforms a relative flow R into an absolute flow A, checking for margins
// Ax = X + Rx , Ay = Y + Ry
// if either
//   - Ax is outside [0,w] or
//   - Ay outside [0,y]
// it returns kPOSITION_INVALID
inline Vec2i getAbsoluteFlow(int x, int y, const Vec2f& flow, int h, int w)
{
	Vec2i result(cvRound(y + flow[1]), cvRound(x + flow[0]));
	if(result[0] >= 0 && result[0] < h && result[1] >= 0 && result[1] < w)
		return result;
	else
		return kPOSITION_INVALID;
}

// Computes the normalized confidence map C between forward flow F and backward flow B
// First, the disntance D at every position (X,Y) is computed :
//     D(X,Y) = ||F(X,Y) - B(X + Fx(X,Y),y + Fy(X,Y))||
// Then, normalized confidence map C is computed :
//     C' = 1 - D / max(D)
Mat1f getFlowConfidence(Mat2f forward_flow, Mat2f backward_flow)
{
	Mat1f distances = Mat1f(forward_flow.rows, forward_flow.cols,-1);

	int h = forward_flow.rows;
	int w = forward_flow.cols;
	float max_distance = -1;

	// Computes the distance between forward and backwards flow
	#pragma omp parallel for
	for(int y = 0; y < h ; ++y)
	{
		for(int x = 0; x < w ; ++x)
		{
			// If there is forward flow for the position F(x,y) 
			Vec2f foward = forward_flow.at<Vec2f>(y, x);
			if(foward[0] != kFLOW_UNKNOWN[0] && foward[1] != kFLOW_UNKNOWN[1])
			{
				Vec2i next_position = getAbsoluteFlow(x, y, foward, h, w);
				if(next_position != kPOSITION_INVALID)
				{
					// If there is backward flow for the refered position B(x + F(x,y).x,y + F(x,y).y)
					Vec2f backward = backward_flow.at<Vec2f>(next_position[0], next_position[1]);
					if(backward[0] != kFLOW_UNKNOWN[0] && backward[1] != kFLOW_UNKNOWN[1])
					{
						// computes the distance
						float distance = (float)norm(foward + backward);

						// Updates the max distance, if required
						if(distance > max_distance)
						{
							#pragma omp critical
							{
								if(distance > max_distance)
									max_distance = distance;
							}
						}

						// Updates the distance map
						distances.at<float>(y, x) = distance;
					}
				}
			}
		}
	}

	// If there is a difference between F and B
	if(max_distance > 0)
	{
		// Computes the normalized confidence map
		#pragma omp parallel for
		for(int y = 0; y < h ; ++y)
		{
			for(int x = 0; x < w ; ++x)
			{
				if(distances.at<float>(y, x) < 0)
				{
					// Unknown flow, C = 0
					distances.at<float>(y, x) = 0;
				}
				else
				{
					// C = 1 - normalized distance
					distances.at<float>(y, x) = (max_distance - distances.at<float>(y, x)) / max_distance;
				}
			}
		}

		return distances;
	}
	else
	{
		// Forward and backwards flow are the same, C = 1
		return Mat1f::ones(forward_flow.rows, forward_flow.cols);
	}
}

// Computes the normalized confidence map C between forward disparity F and backward disparity B
// First, the distance D at every position (X,Y) is computed :
//     D(X,Y) = |F(X,Y) - B(X + F(X,Y),y)|
// then, the normalized confidence map C is computed, using either the maximum value from the map or the optional parameter m :
//     C = 1 - D / Max(m, Max(D))
Mat1f getDisparityConfidence(Mat1f forward_disparity, Mat1f backward_disparity, float max_distance)
{
	Mat1f distances = Mat1f::zeros(forward_disparity.rows, forward_disparity.cols) - 1;

	int h = forward_disparity.rows;
	int w = forward_disparity.cols;
	bool update_max = false;

	if(max_distance < 0)
		update_max = true;

	#pragma omp parallel for
	for(int y = 0; y < h ; ++y)
	{
		for(int x = 0; x < w ; ++x)
		{
			float foward = forward_disparity.at<float>(y, x);
			if(foward != kMOVEMENT_UNKNOWN)
			{
				Vec2i next_position = getAbsoluteFlow(x, y, foward, h, w);
				if(next_position != kPOSITION_INVALID)
				{
					float backward = backward_disparity.at<float>(next_position[0], next_position[1]);
					if(backward != kMOVEMENT_UNKNOWN)
					{
						float distance = abs((float)(foward + backward));
						if(update_max && distance > max_distance)
						{
							#pragma omp critical
							{
								if(distance > max_distance)
									max_distance = distance;
							}
						}
						distances.at<float>(y, x) = distance;
					}
				}
			}
		}
	}

	if(max_distance > 0)
	{
		#pragma omp parallel for
		for(int y = 0; y < h ; ++y)
		{
			for(int x = 0; x < w ; ++x)
			{
				if(distances.at<float>(y, x) < 0)
				{
					distances.at<float>(y, x) = 0;
				}
				else
				{
					distances.at<float>(y, x) = max(0.0f, (max_distance - distances.at<float>(y, x)) / max_distance);
				}
			}
		}

		return distances;
	}
	else
	{
		return Mat1f::ones(forward_disparity.rows, forward_disparity.cols);
	}
}

// Computes a binary confidence map from a normalized confidence map
// To reduce false positives, it applies a one-pixel erosion before thresholding
Mat1f getConfidenceBinary(Mat1f confidence, float vthreshold)
{
	int erosion_size = 1;
	Mat erosion_kernel = getStructuringElement(MORPH_ERODE, Size(2 * erosion_size + 1, 2 * erosion_size + 1), Point(erosion_size, erosion_size));

	Mat1f result = confidence;
	erode(result, result, erosion_kernel);
	threshold(result, result, vthreshold, 1, CV_THRESH_BINARY);
	return result;
}

// Computes the binary confidence map between stereo disparity pairs, taking into account a maximum disparity and a confidence threshold
Mat1f getDisparityConfidenceBinary(Mat1f& disparity, Mat1f& stereo_disparity, float max, float vthreshold)
{
	int erosion_size = 1;
	Mat erosion_kernel = getStructuringElement(MORPH_ERODE, Size(2 * erosion_size + 1, 2 * erosion_size + 1), Point(erosion_size, erosion_size));

	Mat1f confidence = getDisparityConfidence(disparity, stereo_disparity, max);
	erode(confidence, confidence, erosion_kernel);
	threshold(confidence, confidence, vthreshold, 1, CV_THRESH_BINARY);
	return confidence;
}

// Generates a disparity map D from a RGB image I, taking into account in the encoding the maximum amount of disparities m, and if the depth value is inverse,
// Based on the complementary disparity maps provided at http://richardt.name/publications/dcbgrid/supplement/
//     / ! reverse : D(X,Y) = I(X,Y) * m       disparity range: [0, +m]
//     \ reverse:    D(X,Y) = m - I(X,Y) / m   disparity range: [-m, 0]
Mat1f importDisparity(const Mat3f& disparity_render, bool reverse_disparity, float max_disparities, bool reverse_depth)
{
	Mat1f disparity = Mat1f(disparity_render.rows, disparity_render.cols);

	#pragma omp parallel for
	for(int y = 0; y < disparity.rows; y ++)
	{
		for(int x = 0; x < disparity.cols; x ++)
		{
			float dx = min(disparity_render.at<Vec3f>(y, x)[0], 1.0f) * max_disparities;

			if(reverse_depth)
				dx = (max_disparities - dx);

			if(reverse_disparity)
				dx = -dx;

			disparity.at<float>(y, x) = dx;
		}
	}

	return disparity;
}

// Generates a RGB image I from a disparity map D, taking into account in the encoding the maximum amount of disparities m, and if the depth value is inverse
// This function is not inverse to importDisparity, since it was created for different means. The represented disparity maps map [-m,+m] to [0,1]
//     I(X,Y) = 0,5 + D(X,Y) / (2 * m)      disparity range [-m, +m]
Mat3f exportDisparity(const Mat1f& disparity, float max_disparities)
{
	Mat3f disparity_render = Mat3f(disparity.rows, disparity.cols);

	#pragma omp parallel for
	for(int y = 0; y < disparity_render.rows; y++)
	{
		for(int x = 0; x < disparity_render.cols; x++)
		{
			float dx = 0.5f + disparity.at<float>(y, x) / (2 * max_disparities);

			dx = min(max(dx, 0.0f), 1.0f);
			if(dx > -1000) 
				disparity_render.at<Vec3f>(y, x) = Vec3f(dx, dx, dx);
			else
				disparity_render.at<Vec3f>(y, x) = Vec3f(0, 0, 1);
		}
	}

	return disparity_render;
}

// Computes the resulting image R applying the remap operation of the input image I using the disparity map D:
//     R(X,Y) = I(X + D(X,Y), Y)
Mat3f applyDisparity(Mat3f image, const Mat1f& disparity)
{
	return remap<Vec3f>(image, disparity);
}
