// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#pragma once

#include <opencv2/opencv.hpp>


extern const float kMOVEMENT_UNKNOWN;
extern const cv::Vec2f kFLOW_UNKNOWN;

// Computes a binary confidence map from a normalized confidence map
// In order to reduce false positives, it realizes an erosion prior applying the threshold
cv::Mat1f getConfidenceBinary(cv::Mat1f confidence, float vthreshold = 0.95f);

// Computes the normalized confidence map C between forward flow F and backward flow B
// First, the disntance D at every position (X,Y) is computed :
//     D(X,Y) = ||F(X,Y) - B(X + Fx(X,Y),y + Fy(X,Y))||
// Then, normalized confidence map C is computed :
//     C' = 1 - D / max(D)
cv::Mat1f getFlowConfidence(cv::Mat2f forward_flow, cv::Mat2f backward_flow);

// Computes the normalized confidence map C between forward disparity F and backward disparity B
// First, the distance D at every position (X,Y) is computed :
//     D(X,Y) = |F(X,Y) - B(X + F(X,Y),y)|
// then, the normalized confidence map C is computed, using either the maximum value from the map or the optional parameter m :
//     C = 1 - D / Max(m, Max(D))
cv::Mat1f getDisparityConfidence(cv::Mat1f forward_disparity, cv::Mat1f backward_disparity, float max_distance = -1 /*unlimited*/);

// Computes the binary confidence map between stereo disparity pairs, taking into account a maximum disparity and a confidence threshold
cv::Mat1f getDisparityConfidenceBinary(cv::Mat1f& disparity, cv::Mat1f& stereo_disparity, float max_disparity = -1, float threshold = 0.95);


// Computes the resulting image R from the remap operation of the input image I using the optical flow F :
//     (X,Y) = I(X + Fx(X,Y), Y+ Fy(X,Y))
// Instead of using a per pixel operation, it applies an interpolated warping.
template <class TValue>
cv::Mat_<TValue> remap(cv::Mat_<TValue> image, cv::Mat2f flow);

// Computes the resulting image R of the remap operation of the input image I using the disparity map D :
//     R(X,Y) = I(X + D(X,Y), Y)
// Instead of using a per pixel operation, it applies an interpolated warping.
template <class TValue>
cv::Mat_<TValue> remap(cv::Mat_<TValue> image, cv::Mat1f disparity);

// Computes the resulting image R from the remap operation of the input image I using the optical flow F :
//     R(X,Y) = I(X + Fx(X,Y), Y+ Fy(X,Y))
// The results provided are an aproximation, and it's sugested to use the remap
template <class TValue>
cv::Mat_<TValue> applyFlow(cv::Mat_<TValue> image, cv::Mat2f flow, cv::Mat reference = cv::Mat());

// Computes the resulting image R from the remap operation of the input image I using the optical flow F in reverse fashion :
//     R(X + Fx(X,Y), Y+ Fy(X,Y)) = I(X,Y)
template <class TValue>
cv::Mat_<TValue> applyReverseFlow(cv::Mat_<TValue> image, cv::Mat2f flow, cv::Mat reference = cv::Mat());

// Computes the resulting image R of the remap operation of the input image I using the disparity map D :
//     R(X,Y) = I(X + D(X,Y), Y)
cv::Mat3f applyDisparity(cv::Mat3f image, const cv::Mat1f& disparity);

// Generates a disparity map D from a RGB image I, taking into account in the encoding the maximum amount of disparities m, and if the depth value is inverse,
// Based on the complementary disparity maps provided at http://richardt.name/publications/dcbgrid/supplement/
//     / ! reverse : D(X,Y) = I(X,Y) * m        disparity range: [0, +m]
//     \ reverse:    D(X,Y) = m - I(X,Y) / m    disparity range: [-m, 0]
cv::Mat1f importDisparity(const cv::Mat3f& disparity_render, bool reverse_disparity, float max_disparities = 63.75, bool reverse_depth = false);

//	Generates a RGB image I from a disparity map D, taking into account in the encoding the maximum amount of disparities m, and if the depth value is inverse
//  This function is not inverse to importDisparity, since it was created for different means. The represented disparity maps map [-m,+m] to [0,1]
//      I(X,Y) = 0,5 + D(X,Y) / (2 * m)      disparity range [-m, +m]
cv::Mat3f exportDisparity(const cv::Mat1f& disparity, float max_disparities = 63.75);

/////////////////////////////////////////////////////////////////////
// template functions must be in the same file of their headers
/////////////////////////////////////////////////////////////////////

template <class TValue>
cv::Mat_<TValue> applyFlow(cv::Mat_<TValue> image, cv::Mat2f flow, cv::Mat reference)
{
	cv::Mat_<TValue> result;
	if(reference.empty())
		result = cv::Mat_<TValue>::zeros(image.rows, image.cols);
	else
		result = reference.clone();

	#pragma omp parallel for
	for(int y = 0; y < image.rows ; y++)
	{
		for(int x = 0; x < image.cols ; x++)
		{
			cv::Vec2f flow_vector = flow.at<cv::Vec2f>(y, x);
			cv::Vec2i f(cvRound(y + flow_vector[1]), cvRound(x + flow_vector[0]));
			
			if(f[0]>=0 && f[0]< image.rows &&
			   f[1]>=0 && f[1]< image.cols)
			{
				result.template at<TValue>(f[0], f[1]) = image.template at<TValue>(y, x);
			}
		}
	}

	return result;
}

template <class TValue>
cv::Mat_<TValue> applyReverseFlow(cv::Mat_<TValue> image, cv::Mat2f flow, cv::Mat reference)
{
	
	cv::Mat_<TValue> result;
	if(reference.empty())
		result = cv::Mat_<TValue>::zeros(image.rows, image.cols);
	else
		result = reference.clone();

	int h = image.rows;
	int w = image.cols;
	#pragma omp parallel for
	for(int y = 0; y < h ; y++)
	{
		for(int x = 0; x < w ; x++)
		{
			cv::Vec2f flow_vector = flow.at<cv::Vec2f>(y, x);
			cv::Vec2i f(cvRound(y + flow_vector[1]), cvRound(x + flow_vector[0]));
			
			if(f[0]>=0 && f[0]< image.rows &&
			   f[1]>=0 && f[1]< image.cols)
			{
				result.template at<TValue>(y, x) = image.template at<TValue>(f[0], f[1]);
			}
		}
	}

	return result;
}

template <class TValue>
cv::Mat_<TValue> remap(cv::Mat_<TValue> image, cv::Mat2f flow)
{
	cv::Mat2f flow_absolute(image.rows, image.cols);
	for(int y = 0; y < image.rows ; ++y)
	{
		for(int x = 0; x < image.cols;  ++x)
		{
			cv::Vec2f pixel_flow = flow.at<cv::Vec2f>(y, x);
			flow_absolute.at<cv::Vec2f>(y, x) = cv::Vec2f(x + pixel_flow[0], y + pixel_flow[1]);
		}
	}

	cv::Mat_<TValue> result;
	cv::remap(image, result, flow_absolute, cv::Mat(), CV_INTER_LINEAR);
	return result;
}

template <class TValue>
cv::Mat_<TValue> remap(cv::Mat_<TValue> image, cv::Mat1f flow)
{
	cv::Mat2f flow_absolute(image.rows, image.cols);
	for(int y = 0; y < image.rows; y++)
	{
		for(int x = 0; x < image.cols; x++)
		{
			float pixel_flow = flow.at<float>(y, x);
			flow_absolute.at<cv::Vec2f>(y, x) = cv::Vec2f(x + pixel_flow, float(y));
		}
	}

	cv::Mat_<TValue> result;
	cv::remap(image, result, flow_absolute, cv::Mat(), CV_INTER_LINEAR, cv::BORDER_CONSTANT, cv::Scalar(0, 0, 0));
	return result;
}
