// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include <opencv2/opencv.hpp>
#include <fstream>

#include <VideoIO.h>
#include <Metrics.h>
#include <SystemTools.h>
#include <tclap/CmdLine.h>

#include "DeAnaglypher.h"
#include "globals.h"

using namespace cv;
using namespace std;


bool processVideo(Globals& globals);
void computeStatistics(Globals& globals);

int main(int argc, char** argv)
{
	// Set up command line parser TCLAP.
	TCLAP::CmdLine cmd("DeAnaglyph - Tool for reconstructing full-color stereo 3D videos from anaglyph 3D videos.", ' ', "1.0");
	TCLAP::UnlabeledValueArg<string> videoArg(            "video",         "The name of the video (and the directory it is in).", true, "", "video name", cmd);
	TCLAP::UnlabeledValueArg<int>    firstFrameArg(       "first_frame",   "First frame of the video to process.", true, 0, "first frame", cmd);
	TCLAP::UnlabeledValueArg<int>    numFramesArg(        "num_frames",    "Number of frames of the video to process.", true, 0, "number of frames", cmd);
	TCLAP::SwitchArg                 statsArg(       "",  "stats",         "Record additional statistics.", cmd, false);
	TCLAP::SwitchArg                 processArg(     "",  "no-process",    "Don't process the video, e.g. when only statistics (--stats) are to be computed.", cmd, true);
	TCLAP::SwitchArg                 noDispFilterArg("",  "no-disps-tc",   "Disables the temporal consistency filtering of disparity maps.", cmd, false);
	TCLAP::SwitchArg                 loadDispsArg(   "",  "load-disps",    "Load the ground-truth disparity maps instead of computing disparity maps using SIFT flow.", cmd, false);
	TCLAP::SwitchArg                 hasGTDispsArg(  "g", "gt-disps",      "Indicate that ground-truth disparity maps (prefixes TL and TR) exist for this video.", cmd, false);
	TCLAP::ValueArg<int>             maxSizeArg(     "m", "max-size",      "Maximum image size: the images will be resized if their width or height is higher than this value. [default: no downsampling]", false, 0, "pixels", cmd);
	TCLAP::ValueArg<float>           flowScaleArg(   "",  "flow-scale",    "Scale output optical flow visualisations by this factor. [default: 0.03]", false, 0.03f, "factor", cmd);
	TCLAP::ValueArg<float>           maxDisparityArg("",  "max-disparity", "Maximum positive disparity (which corresponds to white in disparity maps). [default: 255./4]", false, 255.f / 4, "value", cmd);
	TCLAP::ValueArg<int>             stepArg(        "s", "step",          "Step size between frames to process. [default: 1]", false, 1, "frames", cmd);
	TCLAP::ValueArg<int>             digitsArg(      "d", "digits",        "Number of digits with which the frame number is formatted. [default: 4]", false, 4, "number", cmd);
	TCLAP::ValueArg<string>          extensionArg(   "e", "ext",           "Filename extension of image files. [default: png]", false, "png", "extension", cmd);
	TCLAP::ValueArg<string>          outputDirArg(   "o", "output-dir",    "Path to the output directory. [default: .]", false, ".", "path", cmd);
	TCLAP::ValueArg<string>          inputDirArg(    "i", "input-dir",     "Path to the directory containing the directory of the video to be loaded. [default: .]", false, ".", "path", cmd);
	cmd.parse(argc, argv);

	// Populate global settings.
	Globals globals;
	globals.video_name  = videoArg.getValue();
	globals.input_path  = inputDirArg.getValue()  + "/";
	globals.output_path = outputDirArg.getValue() + "/";
	globals.extension   = extensionArg.getValue();
	
	globals.has_gt_disparities = hasGTDispsArg.getValue();
	globals.load_disparities   = loadDispsArg.getValue();
	globals.filter_disparities = !noDispFilterArg.getValue();

	globals.num_digits    = digitsArg.getValue();
	globals.start_at      = firstFrameArg.getValue();
	globals.frames        = numFramesArg.getValue();
	globals.step          = stepArg.getValue();
	globals.max_disparity = maxDisparityArg.getValue();
	globals.flow_scale    = flowScaleArg.getValue();

	if(globals.extension[0] != '.')
		globals.extension = string(".") + globals.extension;

	if(maxSizeArg.getValue() > 0)
		globals.max_size = Size(maxSizeArg.getValue(), maxSizeArg.getValue());

	printf("Processing video '%s' (%i frames from #%i with step %i).\n",
		videoArg.getValue().c_str(),
		numFramesArg.getValue(), firstFrameArg.getValue(), stepArg.getValue());
	printf("Using input directory '%s'\n", globals.input_path.c_str());
	printf("Using output directory '%s.'\n", globals.output_path.c_str());

	if(processArg.getValue())
		processVideo(globals);

	if(statsArg.getValue())
		computeStatistics(globals);

	return 0;
}

bool processVideo(Globals& globals)
{
	if(globals.load_disparities && !globals.has_gt_disparities)
	{
		cerr << "Error: loading non-existing ground-truth disparity maps (try --gt-disps)." << endl << endl;
		return false;
	}

	if(!createDirectory(globals.output_path))
	{
		cerr << "Error: output path '" << globals.output_path << "' could not be created." << endl << endl;
		return false;
	}

	VideoIO loader(globals.input_path + globals.video_name + "/", globals.extension, globals.num_digits);
	loader.setMaxSize(globals.max_size);
	
	globals.leftFrames  = filterLeft( loader.loadSequence("L", globals.start_at, globals.frames, globals.step), globals.left_factor);
	globals.rightFrames = filterRight(loader.loadSequence("R", globals.start_at, globals.frames, globals.step));

	// De-anaglyph parameters
	vector<float> luminance_scale(2);
	luminance_scale[0] = 0.2126729f / (0.0721750f + 0.7151522f);
	luminance_scale[1] = 1.0f;

	// Temporal consistency - optical flow and disparity parameters
	TCParameters flow_parameters;
	TCParameters disparity_parameters;
	{
		float sigma_r  = 0.2f;
		float sigma_s  = 2000;
		float sigma_st = 50;

		flow_parameters = TCParameters(sigma_s, sigma_r, 3, sigma_st, sigma_r, 4, 5, sigma_r, sigma_st, TCParameters::kMethod_NC);
		flow_parameters.luminance_scale = luminance_scale;

		if(globals.filter_disparities)
		{
			printf("Filtering disparity maps for temporal consistency.\n");
			disparity_parameters = TCParameters(sigma_s, sigma_r, 6, sigma_st, sigma_r, 4, 5, sigma_st, sigma_r, TCParameters::kMethod_RF);
			disparity_parameters.luminance_scale = luminance_scale;
		}
		else
		{
			printf("Disabled filtering of disparity maps for temporal consistency.\n");
			disparity_parameters = TCParameters::off();
		}
	}

	// Temporal consistency - colorization parameters
	TCParameters additional_information_parameters;
	{
		float sigma_r  = 0.1f;
		float sigma_s  = 200;
		float sigma_st = 100;

		additional_information_parameters = TCParameters(sigma_s, sigma_r, 3, sigma_st, sigma_r, 2, 1, sigma_st,  sigma_r, TCParameters::kMethod_RF);
		additional_information_parameters.luminance_scale = luminance_scale;
	}

	// De-anaglyph
	vector<KeyFrame> leftFlow;
	DeAnaglypher deAnaglypher(globals);

	if(globals.load_disparities)
	{
		printf("Using ground-truth disparity maps.\n\n");
		FrameSequence leftDisparity  = loader.loadSequence("TL", globals.start_at, globals.frames, globals.step);
		FrameSequence rightDisparity = loader.loadSequence("TR", globals.start_at, globals.frames, globals.step);

		leftFlow = deAnaglypher.deAnaglyph(globals.leftFrames, globals.rightFrames, leftDisparity, rightDisparity, flow_parameters, disparity_parameters, additional_information_parameters);
	}
	else
	{
		printf("\n");
		leftFlow = deAnaglypher.deAnaglyph(globals.leftFrames, globals.rightFrames, flow_parameters, disparity_parameters, additional_information_parameters);
	}

	string csv_path = globals.output_path + "stats-process.csv";
	printf("Writing processing statistics to '%s'.\n", csv_path.c_str());

	ofstream stats;		
	stats.open(csv_path.c_str(), ios::out | ios::app);
	stats << "Video;load disparities;filter_disparities;time_filtering;time_color;time_disparity;time_flow;start_at;frames;step" << endl;
	stats << globals.video_name << ';'
		<< globals.load_disparities << ';'
		<< globals.filter_disparities << ';'
		<< globals.time_filtering << ';'
		<< globals.time_color << ';'
		<< globals.time_disparity << ';'
		<< globals.time_flow << ';'
		<< globals.start_at << ';'
		<< globals.frames <<';'
		<< globals.step << endl;
	stats.close();

	return true;
}

// Compares left/right input frames to reconstructed left/right frames.
void computeStatistics(Globals& globals)
{
	if(globals.load_disparities && !globals.has_gt_disparities)
	{
		cerr << "Error: loading non-existing ground-truth disparity maps (try --gt-disps)." << endl << endl;
		return;
	}

	VideoIO inputLoader(globals.input_path + globals.video_name + "/", globals.extension, globals.num_digits);
	inputLoader.setMaxSize(globals.max_size);
	FrameSequence leftInput  = inputLoader.loadSequence("L", globals.start_at, globals.frames, globals.step);
	FrameSequence rightInput = inputLoader.loadSequence("R", globals.start_at, globals.frames, globals.step);

	VideoIO outputLoader(globals.output_path, globals.extension, globals.num_digits);
	outputLoader.setMaxSize(globals.max_size);
	FrameSequence leftOutput  = outputLoader.loadSequence("LO", globals.start_at, globals.frames, globals.step);
	FrameSequence rightOutput = outputLoader.loadSequence("RO", globals.start_at, globals.frames, globals.step);

	string csv_path = globals.output_path + "stats-reconstruction.csv";
	printf("Writing reconstruction statistics to '%s'.\n", csv_path.c_str());

	ofstream stats;
	stats.open(csv_path.c_str(), ios::out /*| ios::app*/);
	stats << "video;channel;frame;load_disparities;filter_disparities;rmsd;psnr;mssim[0];mssim[1];mssim[2]" << endl;

	for(int f = 0; f < leftInput.size(); ++f)
	{
		int rf = f * globals.step + globals.start_at;

		// Left frame
		{
			Scalar mssim = getMSSIM(leftInput[f], leftOutput[f]);
			double rmsd  = getRMSD( leftInput[f], leftOutput[f]);
			double psnr  = getPSNR( leftInput[f], leftOutput[f]);

			stats << globals.video_name << ";left;" << rf << ';' << globals.load_disparities << ';' << globals.filter_disparities << ';' << rmsd << ';' << psnr << ';' << mssim[0] << ';' << mssim[1] << ';' << mssim[2] << endl;
		}

		// Right frame
		{
			Scalar mssim = getMSSIM(rightInput[f], rightOutput[f]);
			double rmsd  = getRMSD( rightInput[f], rightOutput[f]);
			double psnr  = getPSNR( rightInput[f], rightOutput[f]);
			
			stats << globals.video_name << ";right;" << rf << ';' << globals.load_disparities << ';' << globals.filter_disparities << ';' << rmsd << ';' << psnr << ';' << mssim[0] << ';' << mssim[1] << ';' << mssim[2] << endl;
		}
	}

	stats.close();
}