// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include <SIFTflow.h>
#include <FeatureFlow.h>
#include <FlowUtils.h>
#include <MatrixOperations.h>

#include "DeAnaglyph-utils.h"

using namespace cv;
using namespace std;


// Estimates the disparity between two grayscale images using Farneback flow.
Mat1f getDisparityUsingFB(const Mat1f& image1, const Mat1f& image2)
{
	Mat2f disparity_flow;
	calcOpticalFlowFarneback(image1, image2, disparity_flow, 0.5, 3, 16, 3, 5, 1.2, 0);

	vector<Mat1f> flow_channels(2);
	split(disparity_flow, flow_channels);
	return flow_channels[0];
}

// Estimates the disparity between two grayscale images using SIFT flow.
Mat1f getDisparityUsingSiftFlow(const Mat1f& image1, const Mat1f& image2)
{
	Mat2f flow;
	double scale = 1.0;
	{
		// Scale down the input images to speed up the SIFT flow computation, after, scale up the flow.
		Mat1f image1scaled;
		Mat1f image2scaled;
		resize(image1, image1scaled, Size(), scale, scale, INTER_AREA);
		resize(image2, image2scaled, Size(), scale, scale, INTER_AREA);

		// Compute SIFT images.
		int cellsize    = 3;
		int gridspacing = 1;
		bool IsBoundaryIncluded = true;
		SIFTImage sift1 = computeSIFTImage<float>(image1scaled, cellsize, gridspacing, IsBoundaryIncluded);
		SIFTImage sift2 = computeSIFTImage<float>(image2scaled, cellsize, gridspacing, IsBoundaryIncluded);

		// Compute SIFT flow.
		float alpha = 255;
		int d = int(40 * alpha);
		float gamma = 0.005f * alpha;
		int wsize = 5;
		int topwsize = 10;
		int nIterations = 30;
		int nTopIterations = 60;
		int nlevels = 4;
		bool horizontalSearch = true;
		bool verticalSearch   = false; // disabling verticalSearch provides disparity
		Mat2i flowi = computeSIFTFlowFine(sift1, sift2, alpha, d, gamma, wsize, topwsize, nIterations, nTopIterations, nlevels, horizontalSearch, verticalSearch);
				
		// Scale up the computed flow to match the original resolution.
		flowi.convertTo(flow, CV_32FC2);
		resize(flow, flow, Size(image1.cols, image1.rows), 0, 0, INTER_CUBIC);
		flow = flow / scale;
	}

	// Return horizontal component of flow as disparity map.
	vector<Mat1f> flow_channels(2);
	split(flow, flow_channels);
	return flow_channels[0];
}

// Estimates the disparity between two grayscale images using feature matching.
vector<Mat1f> getDisparityUsingFeatures(const Mat1f& image1, const Mat1f& image2)
{
	vector<Point2f> points = calculatePoints(image1);
	Mat1f mask = Mat1f::zeros(image1.rows, image1.cols);
	Mat2f disparity_flow = getFlowForward(image1, image2, points, mask);
	disparity_flow = applyMask<Vec2f>(disparity_flow, mask);
	vector<Mat1f> flow_channels(2);
	split(disparity_flow, flow_channels);
	flow_channels[1] = mask;
	return flow_channels;
}

// Computes the basic binary confidence C, indicating if the disparity points to a valid position
// where 
//   C(x,y) = 1 if x E [0,Width-1], y E [0,Height-1]
//   C(x,y) = 0 otherwise
Mat1f getDisparityConfidence(Mat1f& disparity)
{
	int h = disparity.rows;
	int w = disparity.cols;
	Mat1f confidence(h, w);

	#pragma omp parallel for
	for(int y = 0; y < h; ++y)
	{
		for(int x = 0; x < w; ++x)
		{
			float x1 = x + disparity.at<float>(y, x);

			if(x1 >= 0 && x1 < w)
			{
				confidence.at<float>(y, x) = 1.0f;
			}
			else
			{
				confidence.at<float>(y, x) = 0.0f;
			}
		}
	}

	return confidence;
}

// Incorporates additional information images to keyframes, including a confidence map.
void addAdditionalInformationColor(vector<KeyFrame>& keyframes, FrameSequence& images, vector<Mat1f>& confidence)
{
	size_t n = keyframes.size();

	#pragma omp parallel for
	for(int i = 0; i < n; i++)
	{
		// image propagation
		vector<Mat1f> color_channels(3);
		split(applyMask<Vec3f>(images[i], confidence[i]), color_channels);

		for(int c = 0; c < color_channels.size(); c++)
			color_channels[c] = color_channels[c].clone();

		keyframes[i].additional_information = color_channels;
		keyframes[i].additional_information.push_back(confidence[i].clone());
	}
}

// Merges the previously incorporated additional information, taking into account it is a color map (3 channels)
vector<Mat3f> mergeAdditionalInformationColor(vector<KeyFrame>& keyframes)
{
	size_t n = keyframes.size();
	vector<Mat3f> result(n);

	#pragma omp parallel for
	for(int i = 0; i < n; i++)
	{
		// color reconstruction
		vector<Mat1f> color_channels = keyframes[i].additional_information;
		Mat1f confidence;

		if(color_channels.size() == 4)
		{
			confidence = color_channels[3];
			color_channels.pop_back();
		}
		merge(color_channels, result[i]);
		if(!confidence.empty())
		{
			result[i] = normalize<Vec3f>(result[i], confidence);
		}
	}
	return result;
}

// Deprecated
/*
vector<Mat1f> mergeAdditionalInformationColorConfidence(vector<KeyFrame>& keyframes)
{
	size_t n = keyframes.size();
	vector<Mat1f> result(n);

	for(int i = 0; i < n; i++)
	{
		result[i] = keyframes[i].additional_information[3];
	}

	return result;
}
*/
// Merges the previously incorporated additional information, taking into account it is a disparity map (1 channel).
vector<Mat1f> mergeAdditionalInformationDisparity(vector<KeyFrame>& keyframes)
{
	size_t n = keyframes.size();
	vector<Mat1f> result(n);
	#pragma omp parallel for
	for(int i = 0; i < n; i++)
	{
		//color reconstruction
		//vector<Mat1f> color_channels = keyframes[i].additional_information;
		//merge(color_channels,result[i]);
		//disparity reconstruction
		Mat1f disparity = keyframes[i].additional_information[0];
		if(keyframes[i].additional_information.size() > 1)
		{
			Mat1f mask = keyframes[i].additional_information[1];
			disparity = normalize<float>(disparity, mask);
		}
		result[i] = disparity;
	}
	return result;
}

// Computes an occlusion map based on left-right checking.
vector<Mat1f> calculateOcclusionsUsingDisparity(vector<Mat1f>& disparity, vector<Mat1f>& stereo_disparity, float threshold, float max)
{
	size_t n = disparity.size();
	vector<Mat1f> confidence(n);
	
	#pragma omp parallel for
	for(int i = 0; i < n; i++)
	{
		 confidence[i] = getDisparityConfidenceBinary(disparity[i], stereo_disparity[i], max, threshold);
	}

	return confidence;
}

// Computes an occlusion map based on color consistency.
vector<Mat1f> calculateOcclusionsUsingColor(FrameSequence& frames, FrameSequence& stereo_frames, vector<Mat1f>& disparity, float vthreshold)
{
	size_t n = disparity.size();
	vector<Mat1f> confidence(n);

	int dilation_size = 1;
	int erosion_size  = 1;
	Mat erosion_kernel  = getStructuringElement(MORPH_ERODE,  Size(2 * erosion_size  + 1, 2 * erosion_size  + 1), Point(erosion_size,  erosion_size));
	Mat dilation_kernel = getStructuringElement(MORPH_DILATE, Size(2 * dilation_size + 1, 2 * dilation_size + 1), Point(dilation_size, dilation_size));

	#pragma omp parallel for
	for(int f = 0; f < n; f++)
	{
		confidence[f] = manhattan<float>(frames[f] - applyDisparity(stereo_frames[f], -disparity[f]), true);
		dilate(confidence[f], confidence[f], dilation_kernel);
		//erode(confidence[f], confidence[f], erosion_kernel);
		threshold(confidence[f], confidence[f], vthreshold, 1, CV_THRESH_BINARY_INV);
	}

	return confidence;
}

// Applies an occlusion mask: sets disparity to zero if confidence is very low.
void applyOcclusions(vector<Mat1f>& disparity, vector<Mat1f>& confidence)
{
	size_t n = disparity.size();

	#pragma omp parallel for
	for(int i = 0; i < n; i++)
		disparity[i] = applyMask<float>(disparity[i], confidence[i]);
}

// Warps the images with the given disparity maps.
FrameSequence applyDisparity(FrameSequence& images, vector<Mat1f> disparity)
{
	size_t n = images.size();
	FrameSequence result(n);

	#pragma omp parallel for
	for(int i = 0; i < n; i++)
		result[i] = applyDisparity(images[i], disparity[i]);

	return result;
}

// Merges two sequences depending on per-pixel confidence: use pixel from first sequence if confidence is low, otherwise from second sequence.
FrameSequence mergeUsingConfidence(FrameSequence& frames1, FrameSequence& frames2, vector<Mat1f>& confidence)
{
	size_t n = frames1.size();
	FrameSequence result(n);

	#pragma omp parallel for
	for(int i = 0; i < n; i++)
		result[i] = applyMask<Vec3f>(frames1[i], confidence[i]) + applyMask<Vec3f>(frames2[i], 1 - confidence[i]);

	return result;
}

// Extracts forward disparity maps from a KeyFrame sequence.
vector<Mat1f> getDisparitiesForward(const vector<KeyFrame>& frames)
{
	size_t n = frames.size();
	vector<Mat1f> result(n);

	for(int i = 0; i < n; i++)
		result[i] = frames[i].disparity_forward;

	return result;
}

// Extracts backward disparity maps from a KeyFrame sequence.
vector<Mat1f> getDisparitiesBackward(const vector<KeyFrame>& frames)
{
	size_t n = frames.size();
	vector<Mat1f> result(n);

	for(int i = 0; i < n; i++)
		result[i] = frames[i].disparity_backward;

	return result;
}

// Extracts forward disparity confidence from a KeyFrame sequence.
vector<Mat1f> getDisparityConfidenceForward(const vector<KeyFrame>& frames)
{
	size_t n = frames.size();
	vector<Mat1f> result(n);

	#pragma omp parallel for
	for(int i = 0; i < n; i++)
		result[i] = getConfidenceBinary(frames[i].disparity_confidence_forward, 0.9f);

	return result;
}

// Extracts backward disparity confidence from a KeyFrame sequence.
vector<Mat1f> getDisparityConfidenceBackward(const vector<KeyFrame>& frames)
{
	size_t n = frames.size();
	vector<Mat1f> result(n);

	#pragma omp parallel for
	for(int i = 0; i < n; i++)
		result[i] = getConfidenceBinary(frames[i].disparity_confidence_backward, 0.9f);

	return result;
}

//==== Anaglyph filters ===========================================================================

Mat3f left(const Mat3f& image, Vec3f left_factor)
{
	Mat3f filtered;
	filtered = image.clone();
	vector<Mat1f> channels(3);
	split(filtered, channels);

	channels.at(2) = channels[2] * left_factor[2] + channels[1] * left_factor[1] + channels[0] * left_factor[0]; 
	//channels.at(2) unchanged == full color	
	channels.at(0) = Mat1f::zeros(channels.at(0).rows, channels.at(0).cols);
	channels.at(1) = Mat1f::zeros(channels.at(1).rows, channels.at(1).cols);

	merge(channels, filtered);
	return filtered;
}

Mat3f right(const Mat3f& image)
{
	Mat3f filtered;
	filtered = image.clone();
	vector<Mat1f> channels(3);
	split(filtered, channels);

	channels.at(2) = Mat1f::zeros(channels.at(2).rows, channels.at(2).cols);

	merge(channels, filtered);
	return filtered;
}

Mat3f rightInverse(const Mat3f& image, Vec3f left_factor)
{
	Mat3f filtered;
	filtered = image.clone();
	vector<Mat1f> channels(3);
	split(filtered, channels);

	if(left_factor[1] > 0)
	{
		channels.at(2) = channels[2] - channels[1] * left_factor[1];
	}
	if(left_factor[0] > 0)
	{
		channels.at(2) = channels[2] - channels[0] * left_factor[0];
	}
	channels.at(2) = channels[2] / left_factor[2]; 

	merge(channels, filtered);
	return filtered;
}


FrameSequence filterLeft(FrameSequence frames, Vec3f left_factor)
{
	size_t n = frames.size();
	FrameSequence filtered(n);

	#pragma omp parallel for
	for(int i = 0; i < n; ++i)
		filtered[i] = left(frames[i], left_factor);

	return filtered;
}

FrameSequence filterRight(FrameSequence frames)
{
	size_t n = frames.size();
	FrameSequence filtered(n);

	#pragma omp parallel for
	for(int i = 0; i < n; ++i)
		filtered[i] = right(frames[i]);

	return filtered;
}

FrameSequence filterRightInverse(FrameSequence frames, Vec3f left_factor)
{
	size_t n = frames.size();
	FrameSequence filtered(n);

	#pragma omp parallel for
	for(int i = 0; i < n; ++i)
		filtered[i] = rightInverse(frames[i], left_factor);

	return filtered;
}