// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#pragma once

#include <opencv2/opencv.hpp>
#include <KeyFrame.h>
#include <FeatureFlow.h>

using cv::Mat1f;
using cv::Mat3f;
using cv::Vec3f;
using cv::Size;
using std::string;
using std::vector;


// Estimates the disparity between two grayscale images using Farneback flow.
Mat1f getDisparityUsingFB(
	const Mat1f& image1, // Input source image
	const Mat1f& image2  // Input target image
	);

// Estimates the disparity between two grayscale images using SIFT flow.
Mat1f getDisparityUsingSiftFlow(
	const Mat1f& image1, // Input source image
	const Mat1f& image2  // Input target image
	);

// Estimates the disparity between two grayscale images using feature matching.
vector<Mat1f> getDisparityUsingFeatures(
	const Mat1f& image1, // Input source image
	const Mat1f& image2  // Input target image
	);

// Computes the basic binary confidence C, indicating if the disparity points to a valid position.
Mat1f getDisparityConfidence(
	Mat1f& disparity // Input disparity map
	);


// Incorporates additional information images to keyframes, including a confidence map.
void addAdditionalInformationColor(
	vector<KeyFrame>& keyframes, // Sequence of keyframes to extend
	FrameSequence& images, // Sequence of additional information images to incorporate
	vector<Mat1f>& confidence // Additional information confidence
	);

// Merges the previously incorporated additional information, taking into account it is a color map (3 channels)
vector<Mat3f> mergeAdditionalInformationColor(
	vector<KeyFrame>& keyframes
	);

// Deprecated
//vector<Mat1f> mergeAdditionalInformationColorConfidence(
//	vector<KeyFrame>& keyframes
//	);

// Merges the previously incorporated additional information, taking into account it is a disparity map (1 channel).
vector<Mat1f> mergeAdditionalInformationDisparity(
	vector<KeyFrame>& keyframes
	);


// Computes an occlusion map based on left-right checking.
vector<Mat1f> calculateOcclusionsUsingDisparity(
	vector<Mat1f>& disparity,
	vector<Mat1f>& stereo_disparity,
	float threshold = 0.9f,
	float max = -1
	);

// Computes an occlusion map based on color consistency.
vector<Mat1f> calculateOcclusionsUsingColor(
	FrameSequence& frames,
	FrameSequence& stereo_frames,
	vector<Mat1f>& disparity,
	float vthreshold = 0.1f
	);


// Applies an occlusion mask: sets disparity to zero if confidence is very low.
void applyOcclusions(
	vector<Mat1f>& disparity,
	vector<Mat1f>& confidence
	);

// Warps the images with the given disparity maps.
FrameSequence applyDisparity(
	FrameSequence& images,
	vector<Mat1f> disparity
	);

// Merges two sequences depending on per-pixel confidence: use pixel from first sequence if confidence is low, otherwise from second sequence.
FrameSequence mergeUsingConfidence(
	FrameSequence& frames1,
	FrameSequence& frames2,
	vector<Mat1f>& confidence
	);


// Extracts forward disparity maps from a KeyFrame sequence.
vector<Mat1f> getDisparitiesForward(const vector<KeyFrame>& frames);

// Extracts backward disparity maps from a KeyFrame sequence.
vector<Mat1f> getDisparitiesBackward(const vector<KeyFrame>& frames);

// Extracts forward disparity confidence from a KeyFrame sequence.
vector<Mat1f> getDisparityConfidenceForward(const vector<KeyFrame>& frames);

// Extracts backward disparity confidence from a KeyFrame sequence.
vector<Mat1f> getDisparityConfidenceBackward(const vector<KeyFrame>& frames);


// Returns the left anaglyph channels (by default RED), taking into account the a combination factor
Mat3f left(
	const Mat3f& image, // Input image
	Vec3f left_factor // Combination factor
	);

// Returns the right anaglyph channels (by default CYAN)
Mat3f right(
	const Mat3f& image // Input image
	); 

// Returns the original right channel (by default RED), taking into account that it is the result of combining the original channels with a combination factor 
Mat3f rightInverse(
	const Mat3f& image, // Input image
	Vec3f left_factor // Original combination factor
	);

// Returns a sequence of left channel images, based on the function left(Mat3f,Vec3f)
FrameSequence filterLeft(
	FrameSequence frames, //Sequence of input frames
	Vec3f left_factor // Correction factor
	);

// Returns a sequence of right channel images, based on the function right(Mat3f)
FrameSequence filterRight(
	FrameSequence frames // Sequence of input frames
	);

// Returns a sequence of right channel images, based on the function right(Mat3f), taking into acconunt that the original right channel is a linear combination 
FrameSequence filterRightInverse(
	FrameSequence frames, // Sequence of input frames 
	Vec3f left_factor // Correction factor
	);