// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include <iomanip>
#include <opencv2/opencv.hpp>

#include <RenderFlow.h>
#include <FlowUtils.h>
#include <MatrixOperations.h>
#include <DebugTools.h>
#include <TemporalConsistency.h>
#include <FarnebackFlow.h>

#include "DeAnaglyph-utils.h"
#include "DeAnaglypher.h"

using namespace cv;
using namespace std;


DeAnaglypher::DeAnaglypher(Globals& globals)
	: globals(globals)
	, save_output(true)
{
}

// De-anaglyph from video input only.
vector<KeyFrame> DeAnaglypher::deAnaglyph(FrameSequence& left_frames, FrameSequence& right_frames, TCParameters flow_parameters, TCParameters disparity_parameters, TCParameters additional_information_parameters)
{
	// Compute optical flow for left and right views.
	vector<KeyFrame> left_flow, right_flow;
	computeOpticalFlow(left_frames, right_frames, left_flow, right_flow);

	// Compute disparity maps.
	computeDisparityMaps(left_flow, right_flow);

	// Start the de-anaglyph process.
	return deAnaglyph(left_frames, right_frames, left_flow, right_flow, flow_parameters, disparity_parameters, additional_information_parameters);
}

// De-anaglyph from video + disparity input.
vector<KeyFrame> DeAnaglypher::deAnaglyph(FrameSequence& left_frames, FrameSequence& right_frames, FrameSequence& left_disparity_images, FrameSequence& right_disparity_images, TCParameters flow_parameters, TCParameters disparity_parameters, TCParameters additional_information_parameters)
{
	// Compute optical flow for left and right views.
	vector<KeyFrame> left_flow, right_flow;
	computeOpticalFlow(left_frames, right_frames, left_flow, right_flow);
	
	// Import provided disparity maps.
	importDisparityMaps(left_flow, right_flow, left_disparity_images, right_disparity_images);
	
	// Start the de-anaglyph process.
	return deAnaglyph(left_frames, right_frames, left_flow, right_flow, flow_parameters, disparity_parameters, additional_information_parameters);
}

// De-anaglyph from video, optical flow and disparity maps.
vector<KeyFrame> DeAnaglypher::deAnaglyph(FrameSequence& left_frames, FrameSequence& right_frames, vector<KeyFrame>& left_flow, vector<KeyFrame>& right_flow, TCParameters flow_parameters, TCParameters disparity_parameters, TCParameters additional_information_parameters)
{
	// Temporal consistency: flow + disparity pass.
	cout << "Filtering disparity and optical flow ... ";
	double fTic = tic();
	vector< vector<KeyFrame> > volume(2);
	volume[0] = left_flow;
	volume[1] = right_flow;
	vector< vector<KeyFrame> > filtered_volume = TemporalConsistency(volume, flow_parameters, disparity_parameters, TCParameters::off());
	vector<KeyFrame> result_flow_left  = filtered_volume[0];
	vector<KeyFrame> result_flow_right = filtered_volume[1];
	cout << toc(fTic) << " seconds." << endl;
	
	// Warping input frames using filtered disparities.
	cout << "Warping frames ... ";
	double t = tic();
	vector<Mat1f> left_disparity  = getDisparitiesForward( left_flow);
	vector<Mat1f> right_disparity = getDisparitiesBackward(right_flow);
	vector<Mat1f> left_disparity_filtered  = getDisparitiesForward( result_flow_left);
	vector<Mat1f> right_disparity_filtered = getDisparitiesBackward(result_flow_right);
	FrameSequence left_frames_shifted  = applyDisparity(left_frames,  right_disparity_filtered);
	FrameSequence right_frames_shifted = applyDisparity(right_frames, left_disparity_filtered);
	for(int f = 0; f < left_frames.size(); f++)
	{
		left_frames_shifted[f]  = left_frames_shifted[f]  + right_frames[f];
		right_frames_shifted[f] = right_frames_shifted[f] + left_frames[f];
	}
	cout << toc(t) << " seconds." << endl;

	// Computing confidence from occlusions.
	cout << "Calculating occlusions ... "; t = tic();
	float occlusion_max = 4;
	float occlusion_threshold = 0.25;
	vector<Mat1f> left_disparity_confidence  = calculateOcclusionsUsingDisparity(left_disparity_filtered,  right_disparity_filtered, occlusion_threshold, occlusion_max);
	vector<Mat1f> right_disparity_confidence = calculateOcclusionsUsingDisparity(right_disparity_filtered, left_disparity_filtered,  occlusion_threshold, occlusion_max);
	addAdditionalInformationColor(result_flow_right, left_frames_shifted, right_disparity_confidence);
	addAdditionalInformationColor(result_flow_left, right_frames_shifted, left_disparity_confidence);
	cout << toc(t) << " seconds." << endl;
	globals.time_filtering = toc(fTic);
	
	// Colorization.
	double gTic = tic();
	cout << "Colorizing ... ";
	vector< vector<KeyFrame> > color_volume(2);
	color_volume[0] = result_flow_right;
	color_volume[1] = result_flow_left;
	vector< vector<KeyFrame> > filtered_color_volume = TemporalConsistency(color_volume, TCParameters::off(), TCParameters::off(), additional_information_parameters);
	vector<KeyFrame> result_color_right = filtered_color_volume[0];
	vector<KeyFrame> result_color_left  = filtered_color_volume[1];
	cout << toc(gTic) << " seconds." << endl;

	// Merging colour information.
	cout << "Merging color information ... "; t = tic();
	FrameSequence result_left_frames_shifted  = mergeAdditionalInformationColor(result_color_right);
	FrameSequence result_right_frames_shifted = mergeAdditionalInformationColor(result_color_left);
	FrameSequence result_left_frames_merged   = mergeUsingConfidence(right_frames_shifted, result_right_frames_shifted, left_disparity_confidence);
	FrameSequence result_right_frames_merged  = mergeUsingConfidence(left_frames_shifted,  result_left_frames_shifted,  right_disparity_confidence);
	vector<Mat3f> result_left(left_frames.size());
	vector<Mat3f> result_right(left_frames.size());

	#pragma omp parallel for
	for(int f = 0; f < left_frames.size(); ++f)
	{
		// Use the original left channel (red).
		vector<Mat1f> left_channels(3);
		split(result_left_frames_merged[f], left_channels);
		left_channels[2] = Mat1f::zeros(left_channels[2].rows, left_channels[2].cols);
		merge(left_channels, result_left[f]);
		result_left[f] = rightInverse(result_left[f] + left_frames[f], globals.left_factor);

		// Use the original right channels (green + blue).
		vector<Mat1f> right_channels(3);
		split(result_right_frames_merged[f], right_channels);
		right_channels[0] = Mat1f::zeros(right_channels[0].rows, right_channels[0].cols);
		right_channels[1] = Mat1f::zeros(right_channels[1].rows, right_channels[1].cols);
		merge(right_channels, result_right[f]);
		result_right[f] = rightInverse(result_right[f] + right_frames[f], globals.left_factor);
	}

	cout << toc(t) << " seconds." << endl;
	globals.time_color = toc(gTic);

	if(save_output)
	{
		double t = tic();
		cout << "Exporting results ... ";

		for(int f = 0; f < left_frames.size(); f++)
		{
			int rf = globals.start_at + globals.step * f;

			// File name suffix "####.png".
			ostringstream frame_suffix_builder;
			frame_suffix_builder << setw(globals.num_digits) << setfill('0') << rf << ".png";
			string frame_suffix = frame_suffix_builder.str();

			// Anaglyph input.
			imwrite(globals.output_path + "ANA" + frame_suffix, float2im(left_frames[f] + right_frames[f]));

			// Left and right anaglyph views.
			imwrite(globals.output_path + "LI"  + frame_suffix, float2im(left_frames[f]));
			imwrite(globals.output_path + "RI"  + frame_suffix, float2im(right_frames[f]));

			// Input disparity maps.
			imwrite(globals.output_path + "LDI" + frame_suffix, float2im(exportDisparity(left_disparity[f], globals.max_disparity)));
			imwrite(globals.output_path + "RDI" + frame_suffix, float2im(exportDisparity(right_disparity[f], globals.max_disparity)));

			// Filtered disparity maps.
			imwrite(globals.output_path + "LDO" + frame_suffix, float2im(exportDisparity(left_disparity_filtered[f], globals.max_disparity)));
			imwrite(globals.output_path + "RDO" + frame_suffix, float2im(exportDisparity(right_disparity_filtered[f], globals.max_disparity)));

			// Input optical flow.
			float flow_scale = 1.0f / globals.max_disparity;
			imwrite(globals.output_path + "LIF" + frame_suffix, renderFlow(left_flow[f].flow_forward, Mat3b(), flow_scale));
			imwrite(globals.output_path + "RIF" + frame_suffix, renderFlow(right_flow[f].flow_forward, Mat3b(), flow_scale));

			// Filtered optical flow.
			imwrite(globals.output_path + "LOF" + frame_suffix, renderFlow(result_flow_left[f].flow_forward, Mat3b(), flow_scale));
			imwrite(globals.output_path + "ROF" + frame_suffix, renderFlow(result_flow_right[f].flow_forward, Mat3b(), flow_scale));

			// Disparity confidence.
			imwrite(globals.output_path + "LC"  + frame_suffix, 255 * left_disparity_confidence[f]);
			imwrite(globals.output_path + "RC"  + frame_suffix, 255 * right_disparity_confidence[f]);
			
			// Recombined left and right frames after warping colour channels from other view using filtered disparity.
			imwrite(globals.output_path + "RW"  + frame_suffix, float2im(left_frames_shifted[f]));
			imwrite(globals.output_path + "LW"  + frame_suffix, float2im(right_frames_shifted[f]));

			//// And warped back again to original view (for consistency check).
			//imwrite(globals.output_path + "LWW" + frame_suffix, float2im(applyDisparity(right_frames_shifted[f], -left_disparity[f])));
			//imwrite(globals.output_path + "RWW" + frame_suffix, float2im(applyDisparity(left_frames_shifted[f], -right_disparity[f])));

			// Colorized frames.
			imwrite(globals.output_path + "RF"  + frame_suffix, float2im(result_left_frames_shifted[f]));
			imwrite(globals.output_path + "LF"  + frame_suffix, float2im(result_right_frames_shifted[f]));

			// Merged frames.
			imwrite(globals.output_path + "LM"  + frame_suffix, float2im(result_left_frames_merged[f]));
			imwrite(globals.output_path + "RM"  + frame_suffix, float2im(result_right_frames_merged[f]));
			
			// Resulting left and right frames.
			imwrite(globals.output_path + "LO"  + frame_suffix, float2im(result_left[f]));
			imwrite(globals.output_path + "RO"  + frame_suffix, float2im(result_right[f]));
		}

		cout << toc(t) << " seconds." << endl;
	}

	double tt = globals.time_flow + globals.time_disparity + globals.time_filtering + globals.time_color;
	cout << endl << "Total time: " << tt << " seconds / " << left_frames.size() << " frames = " << tt / left_frames.size() << " seconds / frame." << endl << endl;

	return result_flow_left;
}

// Estimates the optical flow for each channel.
void DeAnaglypher::computeOpticalFlow(FrameSequence& left_frames, FrameSequence& right_frames, vector<KeyFrame>& left_flow, vector<KeyFrame>& right_flow)
{
	cout << "Computing optical flow ... ";
	double vTic = tic();

	//// Optical flow using features.
	//vector<KeyFrame> left_flow  = calculateFlowFeatures(left_frames, 1.0);
	//vector<KeyFrame> right_flow = calculateFlowFeatures(right_frames, 1.0);

	// Farneback optical flow: fast and good enough.
	left_flow  = calculateFlowFB(left_frames);
	right_flow = calculateFlowFB(right_frames);

	globals.time_flow = toc(vTic);
	cout << globals.time_flow << " seconds." << endl;
}

// Estimates the disparity map between the left and right channel of a stereo sequence.
void DeAnaglypher::computeDisparityMaps(vector<KeyFrame>& left_keyframes, vector<KeyFrame>& right_keyframes)
{
	cout << "Computing disparity maps ... ";
	double vTic = tic();

	const size_t n = left_keyframes.size();

	#pragma omp parallel for
	for(int i = 0; i < 2 * n; i++)
	{
		Mat1b left_grey  = grayscale(* left_keyframes[i % n].current);
		Mat1b right_grey = grayscale(*right_keyframes[i % n].current);

		if(i < n)
		{
			Mat1f disparity_l2r = getDisparityUsingSiftFlow(left_grey, right_grey);
			left_keyframes[i].disparity_forward  =  disparity_l2r;
			left_keyframes[i].disparity_backward = -disparity_l2r;
		}
		else
		{
			Mat1f disparity_r2l = getDisparityUsingSiftFlow(right_grey, left_grey);
			right_keyframes[i % n].disparity_backward =  disparity_r2l;
			right_keyframes[i % n].disparity_forward  = -disparity_r2l;
		}
	}

	globals.time_disparity = toc(vTic);
	cout << globals.time_disparity << " seconds." << endl;
}

// Imports the disparity maps for stereo sequences.
void DeAnaglypher::importDisparityMaps(vector<KeyFrame>& left_keyframes, vector<KeyFrame>& right_keyframes, FrameSequence& left_disparities, FrameSequence& right_disparities)
{
	cout << "Importing disparity maps ... ";
	double vTic = tic();

	const size_t n = left_keyframes.size();

	#pragma omp parallel for
	for(int i = 0; i < n; i++)
	{
		Mat1f disparity_l2r = importDisparity(left_disparities[i], false, globals.max_disparity);
		left_keyframes[i].disparity_forward  = -disparity_l2r;
		left_keyframes[i].disparity_backward =  disparity_l2r;

		Mat1f disparity_r2l = importDisparity(right_disparities[i], true, globals.max_disparity);
		right_keyframes[i].disparity_backward = -disparity_r2l;
		right_keyframes[i].disparity_forward  =  disparity_r2l;
	}
	
	globals.time_disparity = toc(vTic);
	cout << globals.time_disparity << " seconds." << endl;
}