// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#pragma once

#include <opencv2/opencv.hpp>

#include <TCParameters.h>

#include "globals.h"
#include "DeAnaglyph-utils.h"

using cv::Mat1f;
using cv::Mat3f;
using cv::Vec3f;
using cv::Size;
using std::string;
using std::vector;


class DeAnaglypher
{
private:
	Globals& globals; // Global configuration parameters
	TCParameters flow_parameters; // Parameters for the Optical flow filtering
	TCParameters disparity_parameters; // Parameters for the disparity filtering
	TCParameters additional_information_parameters; // Parameters for the filtering of additional channels

public:
	bool save_output;
	DeAnaglypher(Globals& globals);

	// De-anaglyph from video input only.
	vector<KeyFrame> deAnaglyph(
		FrameSequence& left_frames, // Left eye sequence
		FrameSequence& right_frames, // Right eye sequence
		TCParameters flow_parameters, // Parameters for the Optical flow filtering
		TCParameters disparity_parameters, // Parameters for the Disparity filtering
		TCParameters additional_information_parameters // Parameters for the filtering of additional channels
		);

	// De-anaglyph from video + disparity input.
	vector<KeyFrame> deAnaglyph(
		FrameSequence& left_frames, // Left eye sequence
		FrameSequence& right_frames, // Right eye sequence
		FrameSequence& left_disparity, // Left eye disparity
		FrameSequence& right_disparity, // Right eye disparity
		TCParameters flow_parameters, // Parameters for the Optical flow filtering
		TCParameters disparity_parameters, // Parameters for the disparity filtering
		TCParameters additional_information_parameters // Parameters for the filtering of additional channels
		);

	// De-anaglyph from video, optical flow and disparity maps.
	vector<KeyFrame> deAnaglyph(
		FrameSequence& left_frames, // Left eye sequence
		FrameSequence& right_frames, // Right eye sequence
		vector<KeyFrame>& left_flow, // Left eye optical flow
		vector<KeyFrame>& right_flow, // Right eye optical flow
		TCParameters flow_parameters, // Parameters for the optical flow filtering
		TCParameters disparity_parameters, // Parameters for the disparity filtering
		TCParameters additional_information_parameters // Parameters for the filtering of additional channels
		);


	// Estimates the optical flow for each channel.
	void computeOpticalFlow(
		FrameSequence& left_frames, // Sequence of left channel frames
		FrameSequence& right_frames, // Sequence of right channel frames
		vector<KeyFrame>& left_flow, // Sequence of per frame optical flow, asociated to the left channel
		vector<KeyFrame>& right_flow // Sequence of per frame optical flow, asociated to the right channel
		);

	// Estimates the disparity map between the left and right channel of a stereo sequence.
	void computeDisparityMaps(
		vector<KeyFrame>& left_keyframes, // Sequence of left keframes, the disparity will be associated to each frame
		vector<KeyFrame>& right_keyframes // Sequence of right keframes, the disparity will be associated to each frame 
		);

	// Imports the disparity maps for stereo sequences.
	void importDisparityMaps(
		vector<KeyFrame>& left_keyframes, // Input left channel sequence
		vector<KeyFrame>& right_keyframes, // Input right channel sequence
		FrameSequence& left_disparities, // Output left channel disparities
		FrameSequence& right_disparities // Output right channel disparities
		);
};