// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

////////////////////////////////////////////////////////////////////////////////
//  This file contains an implementation of the edge-aware filtering approach
//  based on the Domain Transform. Please see the original paper for details:
//
////    Domain Transform for Edge-Aware Image and Video Processing
////    Eduardo S. L. Gastal  and  Manuel M. Oliveira
////    ACM Transactions on Graphics. Volume 30 (2011), Number 4.
////    Proceedings of SIGGRAPH 2011, Article 69.
////
////  Please refer to the publication above if you use this software.
////  For an up-to-date version go to:
////  
////           http://inf.ufrgs.br/~eslgastal/DomainTransform/
////
////  Version 1.0 - August 2011.
//
//  This C++ implementation is equivalent to the MATLAB implementation provided
//  by the original authors.
////////////////////////////////////////////////////////////////////////////////

#pragma once

#define _USE_MATH_DEFINES
#include <opencv2/opencv.hpp>
#include <cmath>
#include <assert.h>

#include <MatrixOperations.h>


// Option for method selection
struct FilteringMethod
{
	enum Value
	{
		NC = 0, // Normalized convolution method
		IC = 1, // Interpolated convolution method
		RF = 2  // Recursive filtering method
	};
};


// Computes the Domain transform image in the given direction
//   - TRef: data type contained in the reference matrix (e.g. Vec3f for float RGB images)
template <class TRef>
inline cv::Mat1f DomainTransform(
	cv::Mat_<TRef> src, // Input image
	float sigma_r,      // Kernel size in space
	float sigma_s,      // Kernel size in range 
	int direction       // Computation direction (kHORIZONTAL|kVERTICAL)
);

// Normalized convolution with optional joint image
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
//   - TRef: data type contained in the reference matrix (e.g. Vec3f for float RGB images)
template <class TValue, class TRef>
cv::Mat_<TValue> NormalizedConvolution2d(
	cv::Mat_<TValue> src, // Input image
	float sigma_s,        // Kernel size in space
	float sigma_r,        // Kernel size in range 
	int iterations = 3,   // Times to iterate
	cv::Mat_<TRef> joint_image = cv::Mat_<TRef>() // [Optional] joint image to compute the transform domain
);

// Normalized convolution with precalculated integrated domain transforms
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
template <class TValue>
cv::Mat_<TValue> NormalizedConvolution2d(
	cv::Mat_<TValue> src, // Input image
	cv::Mat1f ct_H,       // Integrated horizontal domain transform
	cv::Mat1f ct_V,       // Integrated vertical domain transform
	float sigma_s,        // Kernel size in space
	int iterations = 3    // Times to iterate
);

// Normalized convolution with precalculated integrated domain transforms
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
template <class TValue>
cv::Mat_<TValue> NormalizedConvolution1d(
	cv::Mat_<TValue> src, //  Input image 
	cv::Mat1f ct_H,       // Integrated horizontal domain transform
	float sigma_s,        // Kernel size in space
	int iterations = 3    // Times to iterate
);

// Interpolated convolution with optional joint image
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
//   - TRef: data type contained in the reference matrix (e.g. Vec3f for float RGB images)
template <class TValue, class TRef>
cv::Mat_<TValue> InterpolatedConvolution2d(
	cv::Mat_<TValue> src, // Input image
	float sigma_s,        // Kernel size in space
	float sigma_r,        // Kernel size in range 
	int iterations = 3,   // Times to iterate
	cv::Mat_<TRef> joint_image = cv::Mat_<TRef>() // Optional joint image to compute the transform domain
);

// Interpolated convolution with precalculated integrated domain transforms
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
template <class TValue>
cv::Mat_<TValue> InterpolatedConvolution2d(
	cv::Mat_<TValue> src, //  Input image
	cv::Mat1f ct_H,       // Integrated horizontal domain transform
	cv::Mat1f ct_V,       // Integrated vertical domain transform
	float sigma_s,        // Kernel size in space
	int iterations = 3    // Times to iterate
);

// Interpolated convolution with precalculated integrated domain transforms
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
template <class TValue>
cv::Mat_<TValue> InterpolatedConvolution1d(
	cv::Mat_<TValue> src, // Input image
	cv::Mat1f ct_H,       // Integrated horizontal domain transform
	float sigma_s,        // Kernel size in space
	int iterations = 3    // Times to iterate
);

// Recursive filtering with optional joint image
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
//   - TRef: data type contained in the reference matrix (e.g. Vec3f for float RGB images)
template <class TValue, class TRef>
cv::Mat_<TValue> RecursiveFiltering2d(
	cv::Mat_<TValue> src, // Input image
	float sigma_s,        // Kernel size in space
	float sigma_r,        // Kernel size in range 
	int iterations = 3,   // Times to iterate
	cv::Mat_<TRef> joint_image = cv::Mat_<TRef>() // Optional joint image to compute the transform domain
);

// Recursive filtering with precalculated domain
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
template <class TValue>
cv::Mat_<TValue> RecursiveFiltering2d(
	cv::Mat_<TValue> src, // Imput image
	cv::Mat1f dHdx,       // Horizontal Domain transform
	cv::Mat1f dVdy,       // Vertical Domain transform
	float sigma_s,        // Kernel size in space
	int iterations = 3    // Times to iterate
);

// Recursive filtering with optional joint
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
//   - TRef: data type contained in the reference matrix (e.g. Vec3f for float RGB images)
template <class TValue, class TRef>
cv::Mat_<TValue> RecursiveFiltering1d(
	cv::Mat_<TValue> src, // Input image
	float sigma_s,        // Kernel size in space
	float sigma_r,        // Kernel size in range 
	int iterations = 3,   // Times to iterate
	cv::Mat_<TRef> joint_image = cv::Mat_<TRef>() // Optional joint image to compute the transform domain
);

// Recursive filtering with precalculated domain
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
template <class TValue>
cv::Mat_<TValue> RecursiveFiltering1d(
	cv::Mat_<TValue> src, // Input image
	cv::Mat1f dHdx,       // Horizontal Domain transform
	float sigma_s,        // Kernel size in space
	int iterations = 3    // Times to iterate
);


// Obtains a mask from the difference between two images, from the average pixel value.
// NOTE: This method is not optimized
template <class TValue>
cv::Mat1f getMask(const cv::Mat_<TValue>& image1, const cv::Mat_<TValue>&  image2, float min = 0.005f, float max = 1.0);


// Colorization technique, based on the Domain transform filtering
// It propagates a color information (scribbles) over a reference image
//   - TValue: data type contained both in the reference and the scribbles (e.g. Vec3f for float RGB images)
template <class TValue>
cv::Mat_<TValue> colorize(
	cv::Mat_<TValue> reference, // Input image 
	cv::Mat_<TValue> scribbles, // Color information
	cv::Mat1f scribbles_mask,   // Confidence mask of the color information
	float sigma_s = 100,        // Kernel size in space
	float sigma_r = 0.03f,      // Kernel size in range
	int iterations = 3,         // Times to iterate
	FilteringMethod::Value method = FilteringMethod::RF // Domain transform filtering method
);

// General propagation technique, based on the Domain transform filtering
// It propagates information (scribbles) over a reference image
//   - TValue: data type contained in the information matrix (e.g. Vec3f for float RGB images)
//   - TRef: data type contained in the reference matrix (e.g. Vec3f for float RGB images)
template <class TValue, class TRef>
cv::Mat_<TValue> propagate(
	cv::Mat_<TRef> reference,   // Reference image from where to extract the edge information
	cv::Mat_<TValue> scribbles, // Information to propagate
	cv::Mat1f scribbles_mask,   // Information confidence mask
	float sigma_s = 100,        // Kernel size in space
	float sigma_r = 0.03f,      // Kernel size in range
	int iterations = 3,         // Times to iterate
	FilteringMethod::Value method = FilteringMethod::RF // Domain transform filtering method
);

// General propagation technique, based on the Domain transform filtering
// It propagates information (scribbles) over a reference image, from which the domain transform information was precomputed
//   - TValue: data type contained in the input matrix (e.g. Vec3f for float RGB images)
template <class TValue>
cv::Mat_<TValue> propagate(
	cv::Mat_<TValue> scribbles, // Information to propagate
	cv::Mat1f scribbles_mask,   // Information confidence mask
	cv::Mat1f dHdx,             // Precomputed horizontal domain transform
	cv::Mat1f dVdy,             // Precomputed vertical domain transform
	float sigma_s = 100,        // Kernel size in space
	float sigma_r = 0.03f,      // Kernel size in range
	int iterations = 3,         // Times to iterate
	FilteringMethod::Value method = FilteringMethod::RF // Domain transform filtering method
);


/////////////////////////////////////////////////////////////////////
// template functions must be in the same file of their headers
/////////////////////////////////////////////////////////////////////

template <class TValue>
cv::Mat_<TValue> TransformedDomainNCBoxFilter1d(cv::Mat_<TValue> I, cv::Mat1f xform_domain_position, float box_radius);
template <class TValue>
cv::Mat_<TValue> TransformedDomainICBoxFilter1d(cv::Mat_<TValue> I, cv::Mat1f xform_domain_position, float box_radius);
template <class TValue>
cv::Mat_<TValue> TransformedDomainRecursiveFilter2d(cv::Mat_<TValue> I, cv::Mat1f Dx, cv::Mat1f Dy, float sigma);
template <class TValue>
cv::Mat_<TValue> TransformedDomainRecursiveFilter1d(cv::Mat_<TValue> I, cv::Mat1f Dx, float sigma);

inline float SigmaHi(float sigma_H, int i, int N);

//  NC  Domain transform normalized convolution edge-preserving filter.
// 
//  F = NC(img, sigma_s, sigma_r, iterations, joint_image)
//
//  Parameters:
//    src           Input image to be filtered.
//    sigma_s       Filter spatial standard deviation.
//    sigma_r       Filter range standard deviation.
//    iterations	Number of iterations to perform (default: 3).
//    joint_image   Optional image for joint filtering.
//
//
//
//  Based on the reference implementation of the domain transform NC filter
//  described in the paper:
//// 
////    Domain Transform for Edge-Aware Image and Video Processing
////    Eduardo S. L. Gastal  and  Manuel M. Oliveira
////    ACM Transactions on Graphics. Volume 30 (2011), Number 4.
////    Proceedings of SIGGRAPH 2011, Article 69.
////
////  Please refer to the publication above if you use this software. For an
////  up-to-date version go to:
////  
////           http://inf.ufrgs.br/~eslgastal/DomainTransform/
////
////  Version 1.0 - August 2011.
template <class TValue, class TRef>
cv::Mat_<TValue> NormalizedConvolution2d(cv::Mat_<TValue> src, float sigma_s, float sigma_r, int iterations, cv::Mat_<TRef> joint_image)
{
	// Input image.
	cv::Mat_<TValue> I = src;

	// Joint image (optional).
	cv::Mat_<TRef> J = I;
	if (!joint_image.empty())
	{
		// Input and joint images must have equal width and height.
		assert(src.size() == joint_image.size());
		J = joint_image;
	}

	// Compute the domain transform (Equation 11 of the paper).
	cv::Mat1f dHdx = DomainTransform<TRef>(J, sigma_r, sigma_s, kHORIZONTAL);
	cv::Mat1f dVdy = DomainTransform<TRef>(J, sigma_r, sigma_s, kVERTICAL);

	// Integrate the domain transforms.
	cv::Mat1f ct_H = cumsum<float>(dHdx, kHORIZONTAL);
	cv::Mat1f ct_V = cumsum<float>(dVdy, kVERTICAL);

	return NormalizedConvolution2d<TValue>(I, ct_H, ct_V, sigma_s, iterations);
}

template <class TValue>
cv::Mat_<TValue> NormalizedConvolution2d(cv::Mat_<TValue> I, cv::Mat1f ct_H, cv::Mat1f ct_V, float sigma_s, int iterations)
{
	// The vertical pass is performed using a transposed image. It's the most expensive operation on the filtering,
	// but allows to take advantage of the data locality during the iteration, thus providing an overall optimization.
	ct_V = ct_V.t();

	// Perform the filtering.
	////////////////////////////////////////
	int N = iterations;
	cv::Mat_<TValue> F = I;
	float sigma_H = sigma_s;

	for (int i = 0; i < iterations; i++)
	{
		// Compute the sigma value for this iteration (Equation 14 of the paper).
		float sigma_H_i = SigmaHi(sigma_H, i, N);
	
		// Compute the radius of the box filter with the desired variance.
		float box_radius = sqrt(3.0f) * sigma_H_i;
		
		// Horizontal pass
		F = TransformedDomainNCBoxFilter1d<TValue>(F, ct_H, box_radius);
		
		// Vertical pass
		F = F.t();
		F = TransformedDomainNCBoxFilter1d<TValue>(F, ct_V, box_radius);
		F = F.t();
	}

	return F;
}

template <class TValue>
cv::Mat_<TValue> NormalizedConvolution1d(cv::Mat_<TValue> I, cv::Mat1f ct_H, float sigma_s, int iterations)
{
	// Perform the filtering.
	////////////////////////////////////////
	int N = iterations;
	cv::Mat_<TValue> F = I;
	float sigma_H = sigma_s;

	for (int i = 0; i < iterations; i++)
	{
		// Compute the sigma value for this iteration (Equation 14 of the paper).
		float sigma_H_i = SigmaHi(sigma_H, i, N);
	
		// Compute the radius of the box filter with the desired variance.
		float box_radius = sqrt(3.0f) * sigma_H_i;
		
		// Horizontal pass
		F = TransformedDomainNCBoxFilter1d<TValue>(F, ct_H, box_radius);
	}

	return F;
}
//// Box filter normalized convolution in the transformed domain.
////////////////////////////////////////////////////////////////////////////
template <class TValue>
cv::Mat_<TValue> TransformedDomainNCBoxFilter1d(cv::Mat_<TValue> I, cv::Mat1f SAT_dt, float box_radius)
{
	int h = I.rows;
	int w = I.cols;

	cv::Mat_<TValue> SAT = cumsum<TValue>(I, kHORIZONTAL); // SAT = summed area table.
	cv::Mat_<TValue> F = cv::Mat_<TValue>::zeros(h, w);

	#pragma omp parallel for
	for (int y = 0; y < h; y++)
	{
		TValue* F_row = F.template ptr<TValue>(y);
		float* SAT_dt_row = SAT_dt.ptr<float>(y % SAT_dt.rows);
		TValue* SAT_row = SAT.template ptr<TValue>(y);

		// since SATs are monotonic and crescent
		// lower bound position N cannot be before the lower bound position N-1.
		int Il_pos_last = 0;
		for (int x = 0; x < w; x++)
		{
			// Compute the lower and upper limits of the box kernel in the transformed domain.
			float tdl = SAT_dt_row[x + 1] - box_radius;
			float tdu = SAT_dt_row[x + 1] + box_radius;
			
			// TODO: fix comment below
			////search the lower and upper limits on the original domain
			// search the lower bound, starting at the last lower bound
			int Il_pos = findIndexOf<float>(tdl, SAT_dt_row, Il_pos_last, w, 1, 0);
			Il_pos_last = Il_pos;

			// search the upper bound, after the lower bound
			int Iu_pos = findIndexOf<float>(tdu, SAT_dt_row, Il_pos, w, 1, w);
			
			// calculate the average value as area integral / interval
			F_row[x] = (SAT_row[Iu_pos] - SAT_row[Il_pos]) / (Iu_pos - Il_pos);
		}
	}

	return F;
}


//  IC  Domain transform interpolated convolution edge-preserving filter.
/////////////////////////////////////////
// 
//  F = IC(img, sigma_s, sigma_r, iterations, joint_image)
//
//  Parameters:
//  img           Input image to be filtered.
//  sigma_s       Filter spatial standard deviation.
//  sigma_r       Filter range standard deviation.
//  iterations  Number of iterations to perform (default: 3).
//  joint_image     Optional image for joint filtering.
//
//
//
// Based on the reference implementation of the domain transform IC filter
// described in the paper:
//   
//    Domain Transform for Edge-Aware Image and Video Processing
//    Eduardo S. L. Gastal  and  Manuel M. Oliveira
//    ACM Transactions on Graphics. Volume 30 (2011), Number 4.
//    Proceedings of SIGGRAPH 2011, Article 69.
//
//  Please refer to the publication above if you use this software. For an
//  up-to-date version go to:
//
//         http://inf.ufrgs.br/~eslgastal/DomainTransform/
//
//Version 1.0 - August 2011.
template <class TValue, class TRef>
cv::Mat_<TValue> InterpolatedConvolution2d(cv::Mat_<TValue> src, float sigma_s, float sigma_r, int iterations, cv::Mat_<TRef> joint_image)
{
	// Input image.
	cv::Mat_<TValue> I = src;

	// Joint image (optional).
	cv::Mat_<TRef> J = I;
	if (!joint_image.empty())
	{
		assert(src.size() == joint_image.size());
		J = joint_image;
	}

	// Compute the domain transform (Equation 11 of the paper).
	cv::Mat1f dHdx = DomainTransform<TRef>(J, sigma_r, sigma_s, kHORIZONTAL);
	cv::Mat1f dVdy = DomainTransform<TRef>(J, sigma_r, sigma_s, kVERTICAL);

	// Integrate the domain transforms.
	cv::Mat1f ct_H = cumsum<float>(dHdx, kHORIZONTAL);
	cv::Mat1f ct_V = cumsum<float>(dVdy, kVERTICAL);
	
	return InterpolatedConvolution2d<TValue>(I, ct_H, ct_V, sigma_s, iterations);
}

template <class TValue>
cv::Mat_<TValue> InterpolatedConvolution2d(cv::Mat_<TValue> I, cv::Mat1f ct_H, cv::Mat1f ct_V, float sigma_s, int iterations)
{
	// The vertical pass is performed using a transposed image.
	ct_V = ct_V.t();

	// Perform the filtering.
	////////////////////////////////////////
	int N = iterations;
	cv::Mat_<TValue> F = I;
	float sigma_H = sigma_s;

	for (int i = 0; i < iterations; i++)
	{
		// Compute the sigma value for this iteration (Equation 14 of the paper).
		float sigma_H_i = SigmaHi(sigma_H, i, N);
	
		// Compute the radius of the box filter with the desired variance.
		float box_radius = sqrt(3.0f) * sigma_H_i;
		
		// Horizontal pass
		F = TransformedDomainICBoxFilter1d<TValue>(F, ct_H, box_radius);

		// Vertical pass
		F = F.t();
		F = TransformedDomainICBoxFilter1d<TValue>(F, ct_V, box_radius);
		F = F.t();
	}

	return F;
}

template <class TValue>
cv::Mat_<TValue> InterpolatedConvolution1d(cv::Mat_<TValue> I, cv::Mat1f ct_H, float sigma_s, int iterations)
{
	// Perform the filtering.
	////////////////////////////////////////
	int N = iterations;
	cv::Mat_<TValue> F = I;
	float sigma_H = sigma_s;

	for (int i = 0; i < iterations; i++)
	{
		// Compute the sigma value for this iteration (Equation 14 of the paper).
		float sigma_H_i = SigmaHi(sigma_H, i, N);
	
		// Compute the radius of the box filter with the desired variance.
		float box_radius = sqrt(3.0f) * sigma_H_i;
		
		// Horizontal pass
		F = TransformedDomainICBoxFilter1d<TValue>(F, ct_H, box_radius);
	}

	return F;
}

// Box filter interpolated convolution in the transformed domain.
////////////////////////////////////////////////////////////////////////////////
template <class TValue>
cv::Mat_<TValue> TransformedDomainICBoxFilter1d(cv::Mat_<TValue> src, cv::Mat1f SAT_dt, float box_radius)
{
	int h = src.rows;
	int w = src.cols;
	
	// result
	cv::Mat_<TValue> F = cv::Mat_<TValue>::zeros(h, w);

	// add borders to the src image and the Transformed domain's SAT
	cv::Mat_<TValue> I = copyMakeBorder<cv::Mat_<TValue> >(src, 0, 0, 1, 1, cv::BORDER_REPLICATE);
	SAT_dt = copyMakeBorder<cv::Mat1f>(SAT_dt, 0, 0, 0, 1, cv::BORDER_REPLICATE);

	// Pixel values outside the bounds of the image are assumed to equal the
	// nearest pixel border value.    
	#pragma omp parallel for
	for (int y = 0; y < h; y++)
	{
		float* SAT_dt_row = SAT_dt.ptr<float>(y);

		SAT_dt_row[0] = SAT_dt_row[1] - 1.2f * box_radius;
		SAT_dt_row[SAT_dt.cols - 1] = SAT_dt_row[SAT_dt.cols - 2] + 1.2f * box_radius;
	}

	// Original Image's SAT, with additional borders
	cv::Mat_<TValue> SAT = cv::Mat_<TValue>::zeros(h, w + 3);
	#pragma omp parallel for
	for (int y = 0; y < h; y++)
	{
		TValue* I_row = I.template ptr<TValue>(y);
		float* SAT_dt_row = SAT_dt.ptr<float>(y);
		TValue* SAT_row = SAT.template ptr<TValue>(y);

		for (int x = 0; x < w; x++)
		{
			// area calculated with the trapezoid rule
			TValue area = (I_row[x + 1] + I_row[x]) * 0.5f * (SAT_dt_row[x + 1] - SAT_dt_row[x]);
			SAT_row[x + 1] = area + SAT_row[x + 0];
		}
	}

	#pragma omp parallel for
	for (int y = 0; y < h; y++)
	{
		TValue* I_row = I.template ptr<TValue>(y);
		TValue* F_row = F.template ptr<TValue>(y);
		float* SAT_dt_row = SAT_dt.ptr<float>(y);
		TValue* SAT_row = SAT.template ptr<TValue>(y);

		//since SATs are monotonic and crescent
		//lower bound position N cannot be before the lower bound position N-1.
		for (int x = 0; x < w; x++)
		{
			// Compute the lower and upper limits of the box kernel in the transformed domain.
			float td_l = SAT_dt_row[x + 1] - box_radius;
			float td_u = SAT_dt_row[x + 1] + box_radius;
			
			// Compute the box filter using a summed area table. This SAT is built using
			// the area under the graph (in the transformed domain) of the interpolated
			// signal. We use linear interpolation and compute the area using the
			// trapezoidal rule.

			// TODO: fix comment
			////Search for the bounds of the two trapezoids (left and right),
			////which intersection describes the central trapezoid
			//inner bounds (central trapezoid)
			int right_l = findIndexOf<float>(td_l, SAT_dt_row, 0, w, 1);
			int left_u = findIndexOf<float>(td_u, SAT_dt_row, right_l, w, 1, w);
			right_l++;

			//outer bounds
			int left_l = right_l - 1;
			int right_u = left_u + 1;
			
			// Full (center) areas.
			TValue C = SAT_row[left_u] - SAT_row[right_l];

			// Left fractional areas.
			float alpha_l = (td_l - SAT_dt_row[left_l]) / (SAT_dt_row[right_l] - SAT_dt_row[left_l]);
			TValue yi_l = I_row[left_l] + alpha_l * (I_row[right_l] - I_row[left_l]);
			TValue L = 0.5f * (yi_l + I_row[right_l]) * (1 - alpha_l) * (SAT_dt_row[right_l] - SAT_dt_row[left_l]);
		
			// Right fractional areas.
			float alpha_r = (td_u - SAT_dt_row[left_u]) / (SAT_dt_row[right_u] - SAT_dt_row[left_u]);
			TValue yi_r = I_row[left_u] + alpha_r * ( I_row[right_u] - I_row[left_u]);
			TValue R = 0.5f * (yi_r + I_row[left_u]) * (alpha_r) * (SAT_dt_row[right_u] - SAT_dt_row[left_u]);
			
			// Compose the areas
			F_row[x] = (L + C + R) / (2 * box_radius);
		}
	}

	return F;
}

//  RF  Domain transform recursive edge-preserving filter.
//////////////////////////////////////////////
//
//  F = RF(img, sigma_s, sigma_r, iterations, joint_image)
//
//  Parameters:
//    img           Input image to be filtered.
//    sigma_s       Filter spatial standard deviation.
//    sigma_r       Filter range standard deviation.
//    iterations  Number of iterations to perform (default: 3).
//    joint_image     Optional image for joint filtering.
//
//
//  Based on the reference implementation of the domain transform RF filter
//  described in the paper:
//
//    Domain Transform for Edge-Aware Image and Video Processing
//    Eduardo S. L. Gastal  and  Manuel M. Oliveira
//    ACM Transactions on Graphics. Volume 30 (2011), Number 4.
//    Proceedings of SIGGRAPH 2011, Article 69.
//
//  Please refer to the publication above if you use this software. For an
//  up-to-date version go to:
//
//           http://inf.ufrgs.br/~eslgastal/DomainTransform/
//
//  Version 1.0 - August 2011.
template <class TValue, class TRef>
cv::Mat_<TValue> RecursiveFiltering2d(cv::Mat_<TValue> src, float sigma_s, float sigma_r, int iterations, cv::Mat_<TRef> joint_image)
{
	// Input image.
	cv::Mat_<TValue> I = src;

	// Joint image (optional).
	cv::Mat_<TRef> J = I;
	if (!joint_image.empty())
	{
		assert(src.size() == joint_image.size());
		J = joint_image;
	}

	// Compute the domain transform (Equation 11 of the paper).
	cv::Mat1f dHdx = DomainTransform<TRef>(J, sigma_r, sigma_s, kHORIZONTAL);
	cv::Mat1f dVdy = DomainTransform<TRef>(J, sigma_r, sigma_s, kVERTICAL);

	return RecursiveFiltering2d<TValue>(I, dHdx, dVdy, sigma_s, iterations);
}

template <class TValue>
cv::Mat_<TValue> RecursiveFiltering2d(cv::Mat_<TValue> I, cv::Mat1f dHdx, cv::Mat1f dVdy, float sigma_s, int iterations)
{
	// Perform the filtering.
	////////////////////////////////////////
	int N = iterations;
	cv::Mat_<TValue> F = I;
	float sigma_H = sigma_s;

	for (int i = 0; i < iterations; i++)
	{
		// Compute the sigma value for this iteration (Equation 14 of the paper).
		float sigma_H_i =  SigmaHi(sigma_H, i, N);

		// Full pass
		F = TransformedDomainRecursiveFilter2d<TValue>(F, dHdx, dVdy, sigma_H_i); 
	}

	return F;
}

// 1D version
template <class TValue, class TRef>
cv::Mat_<TValue> RecursiveFiltering1d(cv::Mat_<TValue> src, float sigma_s, float sigma_r, int iterations, cv::Mat_<TRef> joint_image)
{
	// Input image.
	cv::Mat_<TValue> I = src;

	// Joint image (optional).
	cv::Mat_<TRef> J = I;
	if (!joint_image.empty())
	{
		assert(src.size() == joint_image.size());
		J = joint_image;
	}

	// Compute the domain transform (Equation 11 of the paper).
	cv::Mat1f dHdx = DomainTransform<TRef>(J, sigma_r, sigma_s, kHORIZONTAL);

	return RecursiveFiltering1d<TValue>(I, dHdx, sigma_s, iterations);
}

template <class TValue>
cv::Mat_<TValue> RecursiveFiltering1d(cv::Mat_<TValue> I, cv::Mat1f dHdx, float sigma_s, int iterations)
{
	// Perform the filtering.
	////////////////////////////////////////
	int N = iterations;
	cv::Mat_<TValue> F = I;
	float sigma_H = sigma_s;

	for (int i = 0; i < iterations; i++)
	{
		// Compute the sigma value for this iteration (Equation 14 of the paper).
		float sigma_H_i =  SigmaHi(sigma_H, i, N);

		// Full pass
		F = TransformedDomainRecursiveFilter1d<TValue>(F, dHdx, sigma_H_i); 
	}

	return F;
}

// Recursive filter.
////////////////////////////////////////
template <class TValue>
 cv::Mat_<TValue> TransformedDomainRecursiveFilter2d(cv::Mat_<TValue> I, cv::Mat1f Dx, cv::Mat1f Dy, float sigma)
 {
	// Feedback coefficient (Appendix of our paper).
	float a = exp(-sqrt(2.0f) / sigma);

	cv::Mat_<TValue> F = I;
	cv::Mat1f Vx = pow(a, Dx);
	int h = F.rows;
	int w = F.cols;

	// Horizontal filter
	#pragma omp parallel for
	for (int y = 0; y < h; y++)
	{
		TValue* Frow = F.template ptr<TValue>(y);
		float* Vrow = Vx.ptr<float>(y);

		// Left -> Right filter.
		for (int x = 1; x < w; x++)
		{
			Frow[x] += Vrow[x] * (Frow[x - 1] - Frow[x]);
		}

		// Right -> Left filter.
		for (int x = w - 2; x >= 0; x--)
		{
			Frow[x] += Vrow[x + 1] * (Frow[x + 1] - Frow[x]);
		}
	}

///TODO <optimize>
	cv::Mat1f Vy = pow(a, Dy);
	// Vertical filter
	#pragma omp parallel for
	for (int x = 0; x < w; x++)
	{	
		// Top -> Bottom filter.
		for (int y = 1; y < h; y++)
		{
			TValue& Fxy = F.template at<TValue>(y, x);
			float Vxy = Vy.at<float>(y, x);

			Fxy += Vxy * ( F.template at<TValue>(y - 1, x) - Fxy);
		}

		// Bottom -> Top filter.
		for (int y = h-2; y >=0 ; y--)
		{
			TValue& Fxy = F.template at<TValue>(y, x);
			float Vxyp1 = Vy.at<float>(y + 1, x);

			Fxy += Vxyp1 * (F.template at<TValue>(y + 1, x) - Fxy);
		}
	}
///TODO </optimize>

	return F;
}

template <class TValue>
cv::Mat_<TValue> TransformedDomainRecursiveFilter1d(cv::Mat_<TValue> I, cv::Mat1f Dx, float sigma)
{
	// Feedback coefficient (Appendix of the paper).
	float a = exp(-sqrt(2.0f) / sigma);

	cv::Mat_<TValue> F = I;
	cv::Mat1f Vx = pow(a, Dx);
	int h = F.rows;
	int w = F.cols;

	// Horizontal filter
	#pragma omp parallel for
	for (int y = 0; y < h; y++)
	{
		TValue* Frow = F.template ptr<TValue>(y);
		float* Vrow = Vx.ptr<float>(y % Vx.rows);
		
		// Left -> Right filter.
		for (int x = 1; x < w; x++)
		{
			Frow[x] += Vrow[x] * (Frow[x - 1] - Frow[x]);
		}

		// Right -> Left filter.
		for (int x = w - 2; x >= 0; x--)
		{
			Frow[x] += Vrow[x + 1] * (Frow[x + 1] - Frow[x]);
		}
	}

	return F;
}

template <class TRef>
inline cv::Mat1f DomainTransform(cv::Mat_<TRef> I, float sigma_r, float sigma_s, int direction)
{
	// Compute the domain transform (Equation 11 of the paper).
	int h = I.rows;
	int w = I.cols;
	int num_joint_channels = I.channels();
	
	if(direction == kHORIZONTAL)
	{
		// Estimate vertical partial derivative using finite
		// differences.
		cv::Mat_<TRef> dIcdx = diff<cv::Mat_<TRef> >(I, kHORIZONTAL);
		dIcdx = abs(dIcdx);

		// take the per channel diff, excluding the last col / row
		// since it has no information
///TODO <optimize>
		std::vector<cv::Mat1f> dIcdx_channels(num_joint_channels);			//TODO vector type based on input
		cv::split(dIcdx.colRange(0, dIcdx.cols-1), dIcdx_channels);
	
		// Compute the l1-norm distance of neighbor pixels.
		cv::Mat1f dIdx = cv::Mat1f::zeros(h, w);
		cv::Mat1f dIdx_range = dIdx.colRange(1, dIdx.cols);

		for (int c = 0; c < num_joint_channels; c++)
		{
			dIdx_range = dIdx_range + dIcdx_channels[c];
		}
///TODO </optimize>

		// Compute the derivatives of the horizontal domain transforms.
		return (1 + sigma_s / sigma_r * dIdx);
	}
	else if(direction == kVERTICAL)
	{
		// Estimate horizontal partial derivative using finite differences.
		cv::Mat_<TRef> dIcdy = diff<cv::Mat_<TRef> >(I, kVERTICAL);
		dIcdy = abs(dIcdy);
///TODO <optimize>
		std::vector<cv::Mat1f> dIcdy_channels(num_joint_channels);
		cv::split(dIcdy.rowRange(0, dIcdy.rows - 1), dIcdy_channels);
		
		// Compute the l1-norm distance of neighbor pixels.
		cv::Mat1f dIdy = cv::Mat1f::zeros(h, w);
		cv::Mat1f dIdy_range = dIdy.rowRange(1, dIdy.rows);
	
		for (int c = 0; c < num_joint_channels; c++)
		{
			dIdy_range = dIdy_range + dIcdy_channels[c];
		}
///TODO </optimize>

		// Compute the derivative of the vertical domain transforms.
		return (1 + sigma_s / sigma_r * dIdy);
	}

	return cv::Mat_<TRef>();
}

// Computes the per-iteration sigma_s
inline float SigmaHi(float sigma_H, int i, int N)
{
	return (float)(sigma_H * sqrt(3.0f) * pow(2.0f, N - (1 + i)) / sqrt(float(pow(4.0f, N) - 1)));
}

template <class TValue, class TRef>
cv::Mat_<TValue> propagate(cv::Mat_<TRef> reference, cv::Mat_<TValue> scribbles, cv::Mat1f scribbles_mask, float sigma_s, float sigma_r, int iterations, FilteringMethod::Value method)
{
	int num_channels = scribbles.channels();

	cv::Mat_<TValue> S; // scribbles
	cv::Mat1f M;        // mask

	if (method == FilteringMethod::RF)
	{
		S = RecursiveFiltering2d<TValue, TRef>(scribbles, sigma_s, sigma_r, iterations, reference);
		M = RecursiveFiltering2d<float, TRef>(scribbles_mask, sigma_s, sigma_r, iterations, reference);
	}
	else if (method == FilteringMethod::IC)
	{
		S = InterpolatedConvolution2d<TValue, TRef>(scribbles, sigma_s, sigma_r, iterations, reference);
		M = InterpolatedConvolution2d<float, TRef>(scribbles_mask, sigma_s, sigma_r, iterations, reference);
	}
	else if (method == FilteringMethod::NC)
	{
		S = NormalizedConvolution2d<TValue, TRef>(scribbles, sigma_s, sigma_r, iterations, reference);
		M = NormalizedConvolution2d<float, TRef>(scribbles_mask, sigma_s, sigma_r, iterations, reference);
	}
	
	// applies the normalization function
	std::vector<cv::Mat> channels(num_channels);
	cv::split(S, channels);
	for(int c = 0; c < num_channels; c++)
	{
		cv::divide(channels[c], M, channels[c]);
	}

	// regenerates the 3 channel color image
	cv::Mat_<TValue> F;
	cv::merge(channels, F);
	return F;
}

template <class TValue>
cv::Mat_<TValue> propagate(cv::Mat_<TValue> scribbles, cv::Mat1f scribbles_mask, cv::Mat1f dHdx, cv::Mat1f dVdy, float sigma_s, float sigma_r, int iterations, FilteringMethod::Value method)
{
	int num_channels = scribbles.channels();

	cv::Mat_<TValue> S;
	cv::Mat1f M;
	if (method == FilteringMethod::RF)
	{
		S = RecursiveFiltering2d<TValue>(scribbles, dHdx, dVdy, sigma_s, iterations);
		M = RecursiveFiltering2d<float>(scribbles_mask, dHdx, dVdy, sigma_s, iterations);
	}
	else
	{
		cv::Mat1f ct_H = cumsum<float>(dHdx, kHORIZONTAL);
		cv::Mat1f ct_V = cumsum<float>(dVdy, kVERTICAL);
		if (method == FilteringMethod::IC)
		{
			S = InterpolatedConvolution2d<TValue>(scribbles, ct_H, ct_V, sigma_s, iterations);
			M = InterpolatedConvolution2d<float>(scribbles_mask, ct_H, ct_V, sigma_s, iterations);
		}
		else if (method == FilteringMethod::NC)
		{
			S = NormalizedConvolution2d<TValue>(scribbles, ct_H, ct_V, sigma_s, iterations);
			M = NormalizedConvolution2d<float>(scribbles_mask, ct_H, ct_V, sigma_s, iterations);
		}
	}
	
	// apply the normalization function
	std::vector<cv::Mat1f> channels(num_channels);
	cv::split(S, channels);
	for(int c = 0; c < num_channels; c++)
	{
		cv::divide(channels[c], M, channels[c]);
	}

	// regenerate the 3 channel color image
	cv::Mat_<TValue> F;
	cv::merge(channels, F);
	return F;
}

template <class TValue>
cv::Mat_<TValue> colorize(cv::Mat_<TValue> reference, cv::Mat_<TValue> scribbles, cv::Mat1f scribbles_mask, float sigma_s, float sigma_r, int iterations, FilteringMethod::Value method)
{
	int num_channels = scribbles.channels();

	// Get the color propagation
	cv::Mat_<TValue> F = propagate<TValue, TValue>(reference, scribbles, scribbles_mask, sigma_s, sigma_r, iterations, method);

	// Combine the colorized chromaticity channels with the luminance of the reference image.
	std::vector<cv::Mat1f> F_channels(num_channels);
	std::vector<cv::Mat1f> reference_channels(num_channels);

	cv::Mat_<TValue> F_YCrCb, reference_YCrCb;
	cv::cvtColor(F, F_YCrCb, CV_BGR2YCrCb);
	cv::cvtColor(reference, reference_YCrCb, CV_BGR2YCrCb);

	cv::split(F_YCrCb, F_channels);
	cv::split(reference_YCrCb, reference_channels);

	F_channels[0] = reference_channels[0]; // copy the reference luminance
	cv::merge(F_channels, F_YCrCb);

	// Convert the result back to BGR.
	cv::cvtColor(F_YCrCb, F, CV_YCrCb2BGR);
	return F;
}

// Obtains a mask from the difference between two images, from the average pixel value.
template <class TValue>
cv::Mat1f getMask(const cv::Mat_<TValue>& image1, const cv::Mat_<TValue>&  image2, float min, float max)
{
	cv::Mat_<TValue> mask = abs(image1 - image2);
	int num_channels = mask.channels();
	cv::Mat1f mask_gray;

	if(num_channels == 1)
	{
		mask_gray = mask.clone();
	}
	else
	{
		std::vector<cv::Mat1f> channels(mask.channels());
		cv::split(mask, channels);
		mask_gray = channels[0].clone();
		for(int c = 1; c < num_channels; c++)
		{
			mask_gray += channels[c];
		}
		mask_gray /= mask.channels();
	}
	cv::threshold(mask_gray, mask_gray, min, max, CV_THRESH_BINARY);

	return mask_gray;
}
