// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include <opencv2/opencv.hpp>

#include <DomainTransform.h>
#include <DebugTools.h>

using namespace cv;
using namespace std;


int sigma_s = 100;
int sigma_r = 3;
int iterations = 3;
int method = 3;

Mat3f reference;
Mat3f scribbles_color;
Mat1f scribbles_gray;
Mat3f result_expected;


void showInfo()
{
	cout << "Simple example of colorization, using the different methods of edge-aware" << endl
		<< "filtering using the domain transform, based on the paper:" << endl
		<< "  Domain Transform for Edge-Aware Image and Video Processing" << endl
		<< "  Eduardo S. L. Gastal and Manuel M. Oliveira" << endl
		<< "  ACM Transactions on Graphics, Volume 30 (2011), Number 4" << endl
		<< endl
		<< "Usage:" << endl 
		<< "  DomainTransformColorizationExample <greyscale image> <scribble image>" << endl
		<< "    [<color image for comparison>]" << endl
		<< endl
		<< "Parameters inside the application:" << endl
		<< "  - Method: 0 = no filtering, 1 = NC, 2 = IC, 3 = RF" << endl
		<< "  - Sigma S: kernel size [1-550]" << endl
		<< "  - Sigma R: variance * 100 [1-50]" << endl
		<< "  - Iterations: [0-6]" << endl
		<< endl;
}

void update()
{
	if(method > 0)
	{
		Mat3f result = colorize<Vec3f>(reference.clone(), scribbles_color.clone(), scribbles_gray.clone(), float(sigma_s), sigma_r / 100.0f, iterations, (FilteringMethod::Value)(method - 1));
		imshow("Result", result);

		if(!result_expected.empty())
		{
			imshow("Expected", result_expected);
			showError(result, result_expected, "Error", 1);
		}
	}
	else
	{
		imshow("Result", reference);
	}
}

void on_trackbar(int, void*)
{
	if(sigma_r == 0)
		sigma_r = 1;

	if(sigma_s == 0)
		sigma_s = 1;

	update();
}


int main(int argc, char** argv)
{
	if(argc <= 2)
	{
		showInfo();
		return -1;
	}

	// Load greyscale image.
	reference = im2float(imread(argv[1]));

	// Load and extract scribbles.
	Mat3f scribbles = im2float(imread(argv[2]));
	scribbles_gray  = getMask<Vec3f>(scribbles, reference);
	scribbles_color = applyMask<Vec3f>(scribbles, scribbles_gray);

	// Load comparison image if provided.
	if(argc > 3)
		result_expected = im2float(imread(argv[3]));

	// Create window.
	namedWindow("Result");
	createTrackbar("Method",     "Result", &method,       3, on_trackbar);
	createTrackbar("Sigma S",    "Result", &sigma_s,    550, on_trackbar);
	createTrackbar("Sigma R",    "Result", &sigma_r,     50, on_trackbar);
	createTrackbar("Iterations", "Result", &iterations,   6, on_trackbar);

	on_trackbar(0, 0);
	waitKey(0);

	return 0;
}
