// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include <ctime>
#include <stdio.h>
#include <opencv2/opencv.hpp>

#include <DomainTransform.h>
#include <Metrics.h>

using namespace cv;
using namespace std;


Mat3b desired_result;
Mat3f I;

int sigma_s = 100;
int sigma_r = 10;
int iterations = 3;
int method = 2;


void showInfo()
{
	cout << "Simple example of the different methods of edge-aware filtering" << endl
		<< "using the domain transform, based on the paper:" << endl
		<< "  Domain Transform for Edge-Aware Image and Video Processing" << endl
		<< "  Eduardo S. L. Gastal and Manuel M. Oliveira" << endl
		<< "  ACM Transactions on Graphics, Volume 30 (2011), Number 4" << endl
		<< endl
		<< "Usage:" << endl 
		<< "  DomainTransformFilteringExample <image> [<image for cmoparison>]" << endl
		<< endl
		<< "Parameters inside the application:" << endl
		<< "  - Method: 0 = no filtering, 1 = NC, 2 = IC, 3 = RF" << endl
		<< "  - Sigma S: kernel size [1-150]" << endl
		<< "  - Sigma R: variance * 100 [1-300]" << endl
		<< "  - Iterations: [0-6]" << endl
		<< endl;
}

void update()
{
	Mat3f result;
	clock_t begin = clock();

	switch(method)
	{
		case 1:
			result = NormalizedConvolution2d<Vec3f>(I.clone(), (float)sigma_s, sigma_r / 100.f, iterations, I);
			printf("NC");
			break;

		case 2:
			result = InterpolatedConvolution2d<Vec3f>(I.clone(), (float)sigma_s, sigma_r / 100.f, iterations, I);
			printf("IC");
			break;

		case 3:
			result = RecursiveFiltering2d<Vec3f>(I.clone(), (float)sigma_s, sigma_r / 100.f, iterations, I);
			printf("RF");
			break;

		default:
			result = I;
			break;
	}

	clock_t end = clock();
	double elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;

	imshow("Result", result);

	if(method > 0)
	{
		printf(" | S=%3i | R=%4.2f | %d iters", sigma_s, sigma_r / 100.f, iterations);
		printf(" | %7.4f s = %7.4f MP/s", elapsed_secs, (I.rows * I.cols / 1.e6f) / elapsed_secs);

		if(!desired_result.empty())
		{
			double mse = getRMSD(float2im(result), desired_result);
			printf(" | RMSE: %6.4f", mse);
		}

		printf("\n");
	}
}

void on_trackbar(int, void*)
{
	if(sigma_r == 0)
		sigma_r = 1;

	if(sigma_s == 0)
		sigma_s = 1;

	update();
}


int main(int argc, char** argv)
{
	if(argc <= 1)
	{
		showInfo();
		return -1;
	}

	I = im2float(imread(argv[1]));
	if(!I.data)
	{
		printf("ERROR: Arg 1 points to an invalid image");
		return -1;
	}

	if(argc > 2)
	{
		desired_result = imread(argv[2]);
		if(!desired_result.data)
		{
			printf("ERROR: Arg 2 points to an invalid image");
			return -1;
		}
	}

	// Create window.
	namedWindow("Result");
	createTrackbar("Method",     "Result", &method,       3, on_trackbar);
	createTrackbar("Sigma S",    "Result", &sigma_s,    150, on_trackbar);
	createTrackbar("Sigma R",    "Result", &sigma_r,    300, on_trackbar);
	createTrackbar("Iterations", "Result", &iterations,   6, on_trackbar);

	on_trackbar(0, 0);
	waitKey(0);

	return 0;
}
