// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include <opencv2/opencv.hpp>

#include <SIFTFlow.h>
#include <RenderFlow.h>
#include <FlowUtils.h>
#include <MatrixOperations.h>

using namespace cv;
using namespace std;


// SIFT descriptors
int p_scale = 10; // percentage
int p_cellsize =3;
int p_gridspacing=1;
bool IsBoundaryIncluded = true;

// SIFT Flow
int p_alpha = 100*255;
int p_d = p_alpha*40;
int p_gamma = int(p_alpha * 0.005);
int p_wsize = 3; 
int p_topwsize = 10; 
int p_nIterations = 30; 
int p_nTopIterations = 60; 
int p_nlevels = 4; 
int p_horizontalSearch = 1; 
int p_verticalSearch = 1;  

// render
int p_flow_scale = 10;

// normalized parameters
double scale = 0.66;
float flow_scale = 0.1f;

// resources
Mat3b image1;
Mat3f image1float;
Mat3b image2;
Mat3f image2float;
Mat2f flow;

SIFTImage sift1;
SIFTImage sift2;


void showInfo()
{
	cout << "Example use of SIFT flow [Liu et al. 2011] based on the original" << endl
		<< "C++/MATLAB implementation." << endl
		<< endl
		<< "Usage:" << endl
		<< "  SiftFlowExample <reference image> <query image>" << endl
		<< endl
		<< "Parameters inside the application:" << endl
		<< "  - SIFT image parameters: image scale, cell size, grid spacing" << endl
		<< "  - SIFT belief propagation parameters" << endl
		<< "  - Flow render parameters" << endl
		<< endl
		<< "For details on the parameters, please refer to the publication:" << endl
		<< "  SIFT Flow: Dense Correspondence across Different Scenes" << endl 
		<< "  Ce Liu, Jenny Yuen, Antonio Torralba, Josef Sivic and William T. Freeman" << endl  
		<< "  IEEE Transactions on Pattern Analysis and Machine Intelligence,"<< endl
		<< "  Volume 33, Number 5, 2011" << endl
		<< "  <http://people.csail.mit.edu/celiu/SIFTflow/>" << endl
		<< endl;
}

void force_positive(int& value, int min = 1)
{
	if(value < min)
		value = min;
}

void on_render_trackbar(int, void*)
{
	force_positive(p_flow_scale);

	flow_scale = p_flow_scale / 1000.0f;
	Mat3b flowrender = renderFlow(flow, Mat3b(), flow_scale);

	imshow("SIFT Flow", flowrender);
	imshow("Input 1: reference image", image1);
	imshow("Image 2: query image", image2);

	Mat3b image1rec = remap<Vec3b>(image2, flow);
	imshow("Query image warped to match reference image", image1rec);
}

void on_siftmatch_trackbar(int, void*)
{
	force_positive(p_alpha);
	force_positive(p_wsize);
	force_positive(p_topwsize);
	force_positive(p_nIterations);
	force_positive(p_nTopIterations);
	force_positive(p_nlevels);

	float alpha = p_alpha / 100.0f;
	p_d = int(alpha * 40);
	float gamma = 0.005f * alpha / 100.0f;
	bool horizontalSearch = p_horizontalSearch == 1; 
	bool verticalSearch = p_verticalSearch == 1;  

	// Compute SIFT Flow.
	Mat2i flowi = computeSIFTFlowFine(sift1, sift2, alpha, p_d, gamma, p_wsize, p_topwsize, p_nIterations, p_nTopIterations, p_nlevels, horizontalSearch, verticalSearch);

	// Scale up the result flow to match the original resolution.
	flowi.convertTo(flow, CV_32FC2);
	resize(flow,flow,Size(image1.cols, image1.rows), 0 , 0, INTER_CUBIC);
	flow = flow / scale;

	// update render
	on_render_trackbar(0, 0);
}

void on_siftimage_trackbar(int, void*)
{
	force_positive(p_scale);
	scale = p_scale / 100.0;
	force_positive(p_cellsize);
	force_positive(p_gridspacing);

	// Scale down the imput images to speed up the SIFT flow computation, and afterwards scale up the flow.
	Mat3f image1scaled;
	Mat3f image2scaled;
	resize(image1float, image1scaled, Size(), scale, scale, INTER_AREA);
	resize(image2float, image2scaled, Size(), scale, scale, INTER_AREA);

	// SIFT descriptors
	sift1 = computeSIFTImage<float>(image1scaled, p_cellsize, p_gridspacing, IsBoundaryIncluded);
	sift2 = computeSIFTImage<float>(image2scaled, p_cellsize, p_gridspacing, IsBoundaryIncluded);

	// update matching
	on_siftmatch_trackbar(0, 0);
}


int main(int argc, char* argv[])
{
	if(argc <= 2)
	{
		showInfo();
		return -1;
	}

	image1 = imread(argv[1]);
	image2 = imread(argv[2]);
	image1float = im2float(image1);
	image2float = im2float(image2);

	// Create Window.
	namedWindow("Parameters", WINDOW_NORMAL);

	// Descriptor parameters.
	createTrackbar("scale %",    "Parameters", &p_scale,      100, on_siftimage_trackbar);
	createTrackbar("cell size",  "Parameters", &p_cellsize,    20, on_siftimage_trackbar);
	createTrackbar("grid space", "Parameters", &p_gridspacing,  5, on_siftimage_trackbar);

	// Matching parameters.
	createTrackbar("alpha*100",  "Parameters", &p_alpha,     255*100*10, on_siftmatch_trackbar);
	//createTrackbar("d",          "Parameters", &p_d,      255*100*10*40, on_siftmatch_trackbar);
	//createTrackbar("Gamma*100",  "Parameters", &p_gamma, 255*100*10*0.005, on_siftmatch_trackbar);
	createTrackbar("wsize",      "Parameters", &p_wsize,             15, on_siftmatch_trackbar);
	createTrackbar("iterations", "Parameters", &p_nIterations,      100, on_siftmatch_trackbar);
	//createTrackbar("top iters.", "Parameters", &p_nTopIterations, 200, on_siftmatch_trackbar);
	createTrackbar("levels",     "Parameters", &p_nlevels,           10, on_siftmatch_trackbar);
	createTrackbar("horizontal", "Parameters", &p_horizontalSearch,   1, on_siftmatch_trackbar);
	createTrackbar("vertical",   "Parameters", &p_verticalSearch,     1, on_siftmatch_trackbar);

	// Flow vizualisation.
	createTrackbar("flow scale", "Parameters", &p_flow_scale,       100, on_render_trackbar);

	on_siftimage_trackbar(0, 0);
	cv::waitKey(0);

	return 0;
}
