#pragma once

#include <opencv2/opencv.hpp>

#include "Image.h"
#include "math.h"
#include "memory.h"
#include <vector>

#ifndef M_PI
#define M_PI 3.1415926535897932384626433
#endif

//typedef Image<uchar> SIFTImage;
typedef cv::Mat SIFTImage;

class ImageFeature
{
public:
	ImageFeature(void);
	~ImageFeature(void);
	
	template<class TI>
	static SIFTImage imSIFT(const cv::Mat& imsrc, int cellSize = 2, int stepSize = 1, bool IsBoundaryIncluded = false, int nBins = 8);

	template <class T1,class T2>
	static void imSIFT(const Image<T1> imsrc, Image<T2>& imsift, const vector<int> cellSizeVect, int stepSize = 1, bool IsBoundaryIncluded = false, int nBins = 8);
};



/// TEMPLATE IMPLEMENTATIONS


template<class TI>
SIFTImage ImageFeature::imSIFT(const cv::Mat& im, int cellSize, int stepSize, bool IsBoundaryIncluded, int nBins)
{
	if(cellSize<=0)
	{
		cout<<"The cell size must be positive!"<<endl;
		return SIFTImage();
	}
	DImage imsrc;
	imsrc.fromMat<TI,double>(im);

	// this parameter controls decay of the gradient energy falls into a bin
	// run SIFT_weightFunc.m to see why alpha = 9 is the best value
	int alpha = 9; 

	int width=imsrc.width(),height=imsrc.height(),nchannels =imsrc.nchannels();
	int nPixels = width*height;
	DImage imdx,imdy;
	// compute the derivatives;
	imsrc.dx(imdx,true);
	imsrc.dy(imdy,true);

	// get the maximum gradient over the channels and estimate the normalized gradient
	DImage magsrc(width,height,nchannels),mag(width,height),gradient(width,height,2);
	#pragma omp parallel for
	for(int i=0;i<nPixels;i++)
	{
		double Max;
		int offset = i*nchannels;
		for(int j = 0;j<nchannels;j++)
			magsrc.pData[offset+j] = sqrt(imdx.pData[offset+j]*imdx.pData[offset+j]+imdy.pData[offset+j]*imdy.pData[offset+j]);
		Max = magsrc.pData[offset];
		if(Max!=0)
		{
			gradient.pData[i*2] = imdx.pData[offset]/Max;
			gradient.pData[i*2+1] = imdy.pData[offset]/Max;
		}		
		for(int j = 1;j<nchannels;j++)
		{
			if(magsrc.pData[offset+j]>Max)
			{
				Max = magsrc.pData[offset+j];
				gradient.pData[i*2] = imdx.pData[offset+j]/Max;
				gradient.pData[i*2+1] = imdy.pData[offset+j]/Max;
			}
		}
		mag.pData[i] = Max;
	}

	// get the pixel-wise energy for each orientation band
	DImage imband(width,height,nBins);
	double theta = M_PI*2/nBins;
	#pragma omp parallel for
	for(int k = 0;k<nBins;k++)
	{
		double _cos,_sin,temp;
		_sin    = sin(theta*k);
		_cos   = cos(theta*k);
		for(int i = 0;i<nPixels; i++)
		{
			temp = __max(gradient.pData[i*2]*_cos + gradient.pData[i*2+1]*_sin,0);
			if(alpha>1)
				temp = pow(temp,alpha);
			imband.pData[i*nBins+k] = temp*mag.pData[i]; 
		}
	}

	// filter out the SIFT feature
	DImage filter(cellSize*2+1,1);
	filter[0] = filter[cellSize+1] = 0.25;
	#pragma omp parallel for
	for(int i = 1;i<cellSize+1;i++)
		filter[i+1] = 1;
	#pragma omp parallel for
	for(int i = cellSize+2;i<cellSize*2+1;i++)
		filter[i] = 0;

	DImage imband_cell;
	imband.imfilter_hv(imband_cell,filter.data(),cellSize,filter.data(),cellSize);

	// allocate buffer for the sift image
	int siftdim = nBins*16;
	int sift_width,sift_height,x_shift=0,y_shift=0;

	sift_width = width/stepSize;
	sift_height = height/stepSize;

	if(IsBoundaryIncluded == false)
	{
		sift_width = (width-4*cellSize)/stepSize;
		sift_height= (height-4*cellSize)/stepSize;
		x_shift = 2*cellSize;
		y_shift = 2*cellSize;
	}

	//if(!imsift.matchDimension(sift_width,sift_height,siftdim))
	//	imsift.allocate(sift_width,sift_height,siftdim);
	UCImage imsift(sift_width,sift_height,siftdim);

	// now sample to get SIFT image
	DImage sift_cell(siftdim,1);
	for(int i=0;i<sift_height;i++)
	{
		for(int j =0;j<sift_width;j++)
		{
			int count = 0;
			#pragma omp parallel for
			for(int ii = -1;ii<=2;ii++)
			{
				for(int jj=-1;jj<=2;jj++)
				{
					int y = __min(__max(y_shift+i*stepSize+ii*cellSize,0),height-1);
					int x = __min(__max(x_shift+j*stepSize+jj*cellSize,0),width-1);
					
					// the following code is the same as the above two for debugging purposes
					//int y = y_shift+i*stepSize+ii*cellSize;
					//int x = x_shift+j*stepSize+jj*cellSize;
					//if (x<0 || x>=width)
					//	x = __min(__max(x,0),width-1);
					//if (y<0 || y>=height)
					//	y= __min(__max(y,0),height-1);
					#pragma omp critical
					{
						memcpy(sift_cell.pData+count*nBins,imband_cell.pData+(y*width+x)*nBins,sizeof(double)*nBins);
						count++;
					}
				}
				// normalize the SIFT descriptor
				double mag = sqrt(sift_cell.norm2());
				int offset = (i*sift_width+j)*siftdim;
				//memcpy(imsift.pData+offset,sift_cell.pData,sizeof(double)*siftdim);
				for(int k = 0;k<siftdim;k++)
					imsift.pData[offset+k] = (unsigned char)__min(sift_cell.pData[k]/(mag+0.01)*255,255);//(unsigned char) __min(sift_cell.pData[k]/mag*512,255);
			}
		}
	}
	//return imsift;
	
	SIFTImage matsift(sift_height, sift_width, CV_MAKETYPE(CV_8U, siftdim));
	#pragma omp parallel for
	for(int y = 0; y < sift_height; ++y)
	{
		for(int x = 0; x < sift_width; ++x)
		{
			int offset = (y * sift_width + x) * siftdim;
			for(int c = 0; c < siftdim; ++c)
			{
				matsift.data[offset + c] = imsift.pData[offset + c];
			}
		}
	}
	return matsift;
	
}
