#include "SIFTFlow.h"

using namespace cv;

Mat3b renderSIFTImage(SIFTImage img)
{
	int h = img.rows;
	int w = img.cols;
	int n = img.channels();
	Mat3b result(h,w);
	for (int y = 0; y < h; y++)
	{
		for (int x = 0; x < w; x++)
		{
			uchar csum = 0;
			for(int c = 0; c < n; ++c) //channel
			{
				csum = max(csum, img.data[y*w*n + x*n + c]);
			}
			//csum = csum/n;
			result.at<Vec3b>(y, x) = Vec3b(csum, csum, csum);
		}
	}
	return result;
}


Mat2i computeSIFTFlowCoarse(SIFTImage& im1, SIFTImage& im2, float alpha, int d, float gamma, int wsize, int nIterations, int nHierarchy)
{
	// output parameters
	double* pEnergyList=NULL;

	BPFlow bpflow;
	bpflow.LoadImages(im1.cols,im1.rows,im1.channels(),im1.data,im2.cols,im2.rows,im2.data);
	
	
	bpflow.setPara(alpha,d);

	bpflow.setHomogeneousMRF(wsize); 

	bpflow.ComputeDataTerm();
	bpflow.ComputeRangeTerm(gamma);
	
	bpflow.MessagePassing(nIterations, nHierarchy, pEnergyList);

	bpflow.ComputeVelocity();

	return bpflow.flow();
}
template <class TValue>
Mat_<TValue> meshgridX(int min, int max, int width, int height)
{
	Mat_<TValue> result(height, width);
	double step = double(max - min) / (width - 1);
	//#pragma omp parallel for
	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < width; ++x)
		{
			result.data[y * width + x] = int(min + step * x);
		}
	}
	return result;
}
template <class TValue>
inline Mat_<TValue> meshgridY(int min, int max, int width, int height)
{
	Mat_<TValue> result(height, width);
	double step = double(max - min) / (height - 1);
	//#pragma omp parallel for
	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < width; ++x)
		{
			result.data[y * width + x] = int(min + step * y);
		}
	}
	return result;
}

inline Mat imresize(Mat im, int width, int height)
{
	Mat result;
	cv::resize(im, result, Size(width, height), 0, 0, INTER_CUBIC);
	return result;
}

inline Mat imresize(Mat im, double scale)
{
	Mat result;
	cv::resize(im, result, Size(), scale, scale, INTER_CUBIC);
	return result;
}

inline Mat GaussianSmoothing(Mat input, float sigma, int size)
{
	Mat result;
	GaussianBlur(input, result, Size(size,size), sigma, sigma);
	return result;
}

struct Pyramid
{
	SIFTImage im1;
	SIFTImage im2;
	Mat1f	xx;
	Mat1f	yy;
};

Mat2i computeSIFTFlowFine(SIFTImage& im1, SIFTImage& im2, float alpha, int d, float gamma, int wsize, int topwsize, int nIterations, int nTopIterations, int nlevels, bool horizontalSearch, bool verticalSearch)
{
	vector<Pyramid> pyrd(nlevels);
	// build the pyramid
	pyrd[0].im1=im1;
	pyrd[0].im2=im2;
	
	for(int i = 1; i < nlevels; ++i)
	{
		float sigma = 0.67f;
		int size = 5;
		pyrd[i].im1 = GaussianSmoothing(pyrd[i-1].im1, sigma, size);
		pyrd[i].im1 = imresize(pyrd[i].im1, 0.5);

		pyrd[i].im2 = GaussianSmoothing(pyrd[i-1].im2, sigma, size);
		pyrd[i].im2 = imresize(pyrd[i].im2, 0.5);

	}
	
	for(int i = 0; i < nlevels; ++i)
	{
		int height = pyrd[i].im1.rows;	int width = pyrd[i].im1.cols;
		//int height2 = pyrd[i].im2.rows;	int width2 = pyrd[i].im2.cols;

		// support for diferent resolutions
		//Mat1f x = meshgridX<float>(0, width-1, width, height);
		//Mat1f y = meshgridY<float>(0, height-1, width, height);
		//pyrd[i].xx = ((x - 1) * (width2 - 1) / (width - 1) - (x - 1));
		//pyrd[i].yy = ((y - 1) * (height2 - 1) / (height - 1) - (y - 1));		
		pyrd[i].xx = Mat1f::zeros(height, width); 
		pyrd[i].yy = Mat1f::zeros(height, width); 
	}

	Mat1f vx;	//it require float precision because resize doesn't work with intergers
	Mat1f vy;	//it require float precision because resize doesn't work with intergers
	Mat2i flow;
	for(int i = nlevels - 1; i >= 0 ; --i)
	{
		int height = pyrd[i].im1.rows;	
		int width = pyrd[i].im1.cols;
		//int height2 = pyrd[i].im2.rows;
		//int width2 = pyrd[i].im2.cols;
		
		Mat1i winSizeX;
		Mat1i winSizeY;
		if (i == nlevels - 1)
		{
			vx = pyrd[i].xx.clone();
			vy = pyrd[i].yy.clone();
		
			winSizeX= Mat1i::ones(height,width) * topwsize;
			winSizeY= Mat1i::ones(height,width) * topwsize;
		}
		else
		{
			vx = pyrd[i].xx + imresize(vx - pyrd[i+1].xx, width, height) * 2;
			vy = pyrd[i].yy + imresize(vy - pyrd[i+1].yy, width, height) * 2;	

			winSizeX = Mat1i::ones(height, width) * (wsize + i);
			winSizeY = Mat1i::ones(height, width) * (wsize + i);
		}
		
		float it_gamma = gamma * pow(2.0f, i);
		Mat1i offsetX;
		Mat1i offsetY;
		vx.convertTo(offsetX, CV_32S);
		vy.convertTo(offsetY, CV_32S);
		int it_nIterations;
		int it_nHierarchy;
		if (i == nlevels - 1)
		{
			it_nIterations = nTopIterations;
			it_nHierarchy = 2;
		}
		else
		{
			it_nIterations = nIterations;
			it_nHierarchy = nlevels - i - 1;		
		}

		// output parameters
		double* pEnergyList = NULL;

		// the main function goes here
		BPFlow bpflow;
		bpflow.LoadImages(pyrd[i].im1.cols, pyrd[i].im1.rows, pyrd[i].im1.channels(),pyrd[i].im1.data, pyrd[i].im2.cols, pyrd[i].im2.rows, pyrd[i].im2.data);
	
		bpflow.setPara(alpha,d);
		bpflow.horizontalSearch = horizontalSearch;
		bpflow.verticalSearch = verticalSearch;

		bpflow.setHomogeneousMRF(wsize);
		bpflow.LoadOffset<int>((int*)offsetX.data, (int*)offsetY.data);
		bpflow.LoadWinSize<int>((int*)winSizeX.data, (int*)winSizeY.data);

		bpflow.ComputeDataTerm();
		bpflow.ComputeRangeTerm(it_gamma);
	
		bpflow.MessagePassing(it_nIterations, it_nHierarchy, pEnergyList);

		bpflow.ComputeVelocity();

		flow = bpflow.flow();

		vector<Mat1i> channels(2);		
		split(flow, channels);

		channels[0].convertTo(vx,CV_32F);
		channels[1].convertTo(vy,CV_32F);
	}
	return flow;
}