#pragma once

#include <opencv2/opencv.hpp>
#include "BPFlow.h"

#include "ImageFeature.h"

//  this file contain the implementation of the SIFT Flow, based on the work of
////	Ce Liu
////	CSAIL, MIT
////	Jan, 2009
////	All rights reserved

template<class TI>
SIFTImage computeSIFTImage(const cv::Mat& imsrc, int cellSize = 2, int stepSize = 1, bool IsBoundaryIncluded = false, int nBins = 8);
// Compute SIFT flow image
//
// Usage:
//
// Input arguments
//   imsrc  --          an RGB or grayscale image (either uint8 or double) 
//   cellSize -- (default: 3) a vector of cell size. If a scale is input, then the SIFT descriptor of one cell size is
//                        computed; if a vector is input, then multi-scale SIFT is computed and concatenated. 
//   stepSize -- (default: 1) a scale of step size in sampling image grid. If an integer larger than 1 is input, then
//                       sparse sampling of image grid is performed
//  IsBoundary-- (default: true) a boolean variable indicating whether boundary is included.
//
// Output
//   sift --         an image (with multiple channels, typicallly 128) of datatype UINT8 despite the type of the input.
//                       The maximum element wise value of sift is 255. This datatype is consistent with the byte-based
//                       SIFT flow algorithm
//

cv::Mat2i computeSIFTFlowCoarse(SIFTImage& im1, SIFTImage& im2, float alpha = 0.01, int d =  1, float gamma = 0.001, int wsize = 5, int nIterations = 40, int nHierachy = 2);
cv::Mat2i computeSIFTFlowFine(SIFTImage& im1, SIFTImage& im2, float alpha = 255, int d = 255*40, float gamma = 0.005 * 255, int wsize = 3, int topwsize = 10, int nIterations = 30, int nTopIterations = 60, int nlevels = 4, bool horizontalSearch = true, bool verticalSearch = true);
//	Compute Sift flow (coarse / fine):
//
//		provides the matching of the descriptors using belief propagation. Two alternatives:
//		* Coarse matching: one pass of the belief propagation algorithm
//		* fine matching: multiple levels of pyramid-based belief propagation
//
//  Usage:
// 
//  ** Input arguments **
// 
//  Arguments "im1" and "im2" are SIFT images. SIFT flow is estimated from im1 to im2. 
// 
//  each element is defined as (the number in the parenthesis is the defaut value)
// 
//        alpha:        (0.01) the weight of the truncated L1-norm regularization on the flow
//        d:            (1) the threshold of the truncation
//        gamma:        (0.001) the weight of the magnitude of the flow
//        wsize:        (5)  the half size of the search window 
//        nIterations:  (40) the number of iterations
//        nHierarchy:   (2)  the number of hierarchies in the efficient BP implementation
// 
//  Notice that the two images are NOT required to have the same dimension. But certainly you
//  wouldn't expect to get anything meaningful if their dimensions are too different.
// 
// 
//  ** Output argument **
//  
//  the output is the integer flow, a [h x w x 2] matrix where the first channel is the horizontal component
//	and the second channel the vertical component.
//

// debug
cv::Mat3b renderSIFTImage(SIFTImage img);

// TEMPLATE IMPLEMENTATIONS

template<class TI>
SIFTImage computeSIFTImage(const cv::Mat& imsrc, int cellSize, int stepSize,bool IsBoundaryIncluded, int nBins)
{
	return ImageFeature::imSIFT<TI>(imsrc, cellSize, stepSize, IsBoundaryIncluded, nBins);
}