// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include <opencv2/opencv.hpp>
#include <opencv2/nonfree/features2d.hpp>

#include <FlowUtils.h>
#include <MatrixOperations.h>

#include "SequenceFilter.h"
#include "FeatureFlow.h"

using namespace cv;
using namespace std;


std::vector<KeyPoint> computeFeatures(const Mat1b& img_gray)
{
	SiftFeatureDetector detector(0, 3, 0.00001, 1000);
	vector<KeyPoint> features;
	detector.detect(img_gray, features);
	return features;
}

Mat2f getFlowForward(Mat1b image_current, Mat1b image_next, vector<Point2f> points_current, Mat1f mask)
{
	Mat2f flow = Mat2f(image_current.rows, image_current.cols, kFLOW_UNKNOWN);

	Size winSize(63, 63);
	vector<uchar> status;
	vector<float> err;
	TermCriteria termcrit(TermCriteria::COUNT | TermCriteria::EPS, 20, 0.03);

	vector<Point2f> points_next;
	calcOpticalFlowPyrLK(image_current, image_next, points_current, points_next, status, err, winSize, 3, termcrit, 0, 0.001);

	#pragma omp parallel for
	for(int p = 0; p < points_current.size(); p++)
	{
		if(status[p])
		{
			Point2f match_current = points_current[p];
			Point2f match_next    = points_next[p];
			Vec2i pos(cvRound(match_current.y), cvRound(match_current.x));
			flow.at<Vec2f>(pos[0], pos[1]) = (match_next - match_current);

			if(!mask.empty())
				mask.at<float>(pos[0], pos[1]) = 1.0f;
		}
	}

	return flow;
}

vector<Point2f> calculateCorners(const Mat1b& image_grayscale)
{
	// Apply corner detection
	vector<Point2f> corners;
	goodFeaturesToTrack(image_grayscale, corners, 5000, 0.001, 3, Mat(), 3, false, 0.04);
	return corners;
}

vector<Point2f> calculatePoints(const Mat1b& image_grayscale, float acceptance)
{
	acceptance = max(acceptance, 1.0f);

	// Initialize the points with the corners
	vector<Point2f> points = calculateCorners(image_grayscale);
	points.resize(size_t(points.size() * acceptance));

	// add the feature coordinates to the interesting points
	vector<KeyPoint> features = computeFeatures(image_grayscale);
	for(int f = 0; f < features.size() * acceptance; ++f)
	{
		points.push_back(features[f].pt);
	}

	return points;
}

vector<KeyFrame> calculateFlowFeatures(FrameSequence& frames, float aceptance)
{
	size_t n = frames.size();
	vector<KeyFrame> keyframes(n);

	// Precalculate the information that will be reused.
	vector<Mat1b> gray = calculateGrayscale(frames);

	// Calculate every keyframe
	#pragma omp parallel for
	for(int i = 0; i < n; i++)
	{
		Mat3f& frame = frames[i];
		
		KeyFrame flow;
		flow.number = i;
		flow.current = &frame;
		vector<Point2f> points = calculatePoints(gray[i], aceptance);

		if(i == 0)
		{
			flow.previous = NULL;
			flow.flow_backward = Mat2f::zeros(frame.rows, frame.cols);
		}
		else
		{
			// Relation with the previous frame
			flow.previous = &(frames[i - 1]);
			flow.flow_backward = getFlowForward(gray[i], gray[i - 1], points);
		}

		if(i + 1 < n)
		{
			// Relation with the next frame
			flow.next = &(frames[i + 1]);
			flow.flow_forward = getFlowForward(gray[i], gray[i + 1], points);
		}
		else
		{
			flow.next = NULL;
			flow.flow_forward = Mat2f::zeros(frame.rows, frame.cols);
		}

		keyframes[i] = flow;
	}

	return keyframes;
}
