// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#include <DomainTransform.h>
#include <MatrixOperations.h>

#include "PathKernel.h"

using namespace cv;
using namespace std;


PathKernel::PathKernel(TCParameters* flow_parameters, TCParameters* disparity_parameters, TCParameters* additional_information_parameters)
	: flow_parameters(flow_parameters)
	, disparity_parameters(disparity_parameters)
	, additional_information_parameters(additional_information_parameters)
{
}

PathKernel::PathKernel()
{
}

void PathKernel::init(TCParameters* flow_parameters, TCParameters* disparity_parameters, TCParameters* additional_information_parameters)
{
	this->flow_parameters = flow_parameters;
	this->disparity_parameters = disparity_parameters;
	this->additional_information_parameters = additional_information_parameters; 
}

void PathKernel::updateAdditionalInformationPath(KeyFrameTC& keyframe, int y, int x)
{
	if(!keyframe.additional_information.empty())
	{
		size_t n = keyframe.additional_information.size();
		if (additional_information.empty())
		{
			additional_information = vector< vector<float*> >(n);
		}

		for(int a = 0; a < n; ++a)
		{
			additional_information[a].push_back(&keyframe.additional_information[a].at<float>(y, x));
		}
	}
}

void PathKernel::updateDisparityPath(KeyFrameTC& keyframe, int y, int x)
{
	disparity_backward.push_back(&keyframe.disparity_backward.at<float>(y, x));
	disparity_forward.push_back( &keyframe.disparity_forward.at <float>(y, x));
	disparity_confidence_backward.push_back(&keyframe.disparity_confidence_backward.at<float>(y, x));
	disparity_confidence_forward.push_back( &keyframe.disparity_confidence_forward.at< float>(y, x));
}

void PathKernel::updateFlowPath(KeyFrameTC& keyframe, int y, int x)
{
	flow_backward.push_back(&keyframe.flow_backward.at<Vec2f>(y, x));
	flow_forward.push_back( &keyframe.flow_forward.at< Vec2f>(y, x));
	flow_confidence_backward.push_back(&keyframe.flow_confidence_backward.at<float>(y, x));
	flow_confidence_forward.push_back( &keyframe.flow_confidence_forward.at< float>(y, x));
}

void PathKernel::updatePath(KeyFrameTC& keyframe, int y, int x)
{
	position.push_back(Vec2i(y, x));
	color.push_back(&keyframe.current->at<Vec3f>(y, x));

	if(flow_parameters->filter)
	{
		updateFlowPath(keyframe, y, x);
	}

	if(disparity_parameters->filter)
	{
		updateDisparityPath(keyframe, y, x);
	}

	if(additional_information_parameters->filter)
	{
		updateAdditionalInformationPath(keyframe, y, x);
	}
}

void PathKernel::apply(int filter_mode, int view)
{
	size_t active_path_size = color.size();

	// Check that the path size is long enough to be filtered
	if(active_path_size > 1)
	{
		Mat3f path_color(1, (int)active_path_size);
		for(int p = 0; p < active_path_size; ++p)
		{
			path_color.at<Vec3f>(0, p) = *color[p];
		}

		if(flow_parameters->filter)
		{
			Mat2f path_flow(2, (int)active_path_size);
			Mat1f path_confidence(2, (int)active_path_size);
			for(int p = 0; p < active_path_size; ++p)
			{
				path_flow.at<Vec2f>(0, p) = *flow_forward[p];
				path_flow.at<Vec2f>(1, p) = *flow_backward[p];
				path_confidence.at<float>(0, p) = *flow_confidence_forward[p];
				path_confidence.at<float>(1, p) = *flow_confidence_backward[p];
			}
			// Filter flow and confidence (both directions)

			// Precalculate the joint image's (path colors) domain transform
			float luminance_scale = 1.0f;
			if(!flow_parameters->luminance_scale.empty())
			{
				luminance_scale = flow_parameters->luminance_scale[view];
			}
			Mat1f dPdt = DomainTransform<Vec3f>(path_color, flow_parameters->sigma_rt * luminance_scale, flow_parameters->sigma_st, kHORIZONTAL);

			// Apply the right filter
			Mat2f filtered_path_flow;
			Mat1f filtered_path_confidence;
			if(flow_parameters->filter_method == TCParameters::kMethod_RF)
			{
				filtered_path_flow       = RecursiveFiltering1d<Vec2f>(path_flow,       dPdt, flow_parameters->sigma_st, flow_parameters->iterations);
				filtered_path_confidence = RecursiveFiltering1d<float>(path_confidence, dPdt, flow_parameters->sigma_st, flow_parameters->iterations);
			}
			else
			{
				Mat1f SAT_P = cumsum<float>(dPdt, kHORIZONTAL);
				if(flow_parameters->filter_method == TCParameters::kMethod_NC)
				{
					filtered_path_flow       = NormalizedConvolution1d<Vec2f>(path_flow,       SAT_P, flow_parameters->sigma_st, flow_parameters->iterations);
					filtered_path_confidence = NormalizedConvolution1d<float>(path_confidence, SAT_P, flow_parameters->sigma_st, flow_parameters->iterations);
				}
				else
				{
					filtered_path_flow       = InterpolatedConvolution1d<Vec2f>(path_flow,       SAT_P, flow_parameters->sigma_st, flow_parameters->iterations);
					filtered_path_confidence = InterpolatedConvolution1d<float>(path_confidence, SAT_P, flow_parameters->sigma_st, flow_parameters->iterations);
				}
				
			}

			// Update the values
			for(int p = 0; p < active_path_size; ++p)
			{
				*flow_forward[p]  = filtered_path_flow.at<Vec2f>(0,p);
				*flow_backward[p] = filtered_path_flow.at<Vec2f>(1,p);
				*flow_confidence_forward[p]  = filtered_path_confidence.at<float>(0,p);
				*flow_confidence_backward[p] = filtered_path_confidence.at<float>(1,p);
			}
		}

		if(disparity_parameters->filter)
		{
			float sigma_rt;
			float sigma_st;
			int iterations = 1;

			if(filter_mode == kFilterTime)
			{
				sigma_rt = disparity_parameters->sigma_rt;
				sigma_st = disparity_parameters->sigma_st;
			}
			else
			{
				sigma_rt = disparity_parameters->sigma_rv;
				sigma_st = disparity_parameters->sigma_sv;
			}

			float luminance_scale = 1.0f;
			if(!disparity_parameters->luminance_scale.empty())
			{
				luminance_scale = disparity_parameters->luminance_scale[view];
			}
			sigma_rt = sigma_rt * luminance_scale;

			Mat1f path(4, (int)active_path_size);
			for(int p = 0; p < active_path_size; ++p)
			{
				path.at<float>(0, p) = *disparity_forward[p];
				path.at<float>(1, p) = *disparity_backward[p];
				path.at<float>(2, p) = *disparity_confidence_forward[p];
				path.at<float>(3, p) = *disparity_confidence_backward[p];
			}

			// Precalculate the joint image's (path colors) domain transform
			Mat1f dPdt = DomainTransform<Vec3f>(path_color, sigma_rt, sigma_st, kHORIZONTAL);

			// Apply the right filter
			Mat1f filtered_path;
			if(disparity_parameters->filter_method == TCParameters::kMethod_RF)
			{
				filtered_path = RecursiveFiltering1d<float>(path, dPdt, sigma_st, iterations);
			}
			else
			{
				Mat1f SAT_P = cumsum<float>(dPdt, kHORIZONTAL);
				if(disparity_parameters->filter_method == TCParameters::kMethod_NC)
				{
					filtered_path = NormalizedConvolution1d<float>(path, SAT_P, sigma_st, iterations);
				}
				else
				{
					filtered_path = InterpolatedConvolution1d<float>(path, SAT_P, sigma_st, iterations);
				}
			}

			// Update the values
			for(int p = 0; p < active_path_size; ++p)
			{
				*disparity_forward[p]  = filtered_path.at<float>(0,p);
				*disparity_backward[p] = filtered_path.at<float>(1,p);
				*disparity_confidence_forward[p]  = filtered_path.at<float>(2,p);
				*disparity_confidence_backward[p] = filtered_path.at<float>(3,p);
			}
		}

		// If there is additional information, filter every channel by separate
		if(filter_mode == kFilterTime && additional_information_parameters->filter && !additional_information.empty())
		{
			float luminance_scale = 1.0f;
			if(!additional_information_parameters->luminance_scale.empty())
			{
				luminance_scale = additional_information_parameters->luminance_scale[view];
			}
			float sigma_rt = additional_information_parameters->sigma_rt * luminance_scale;
			
			// Precalculate the joint image's (path colors) domain transform.
			Mat1f dPdt = DomainTransform<Vec3f>(path_color, sigma_rt, additional_information_parameters->sigma_st, kHORIZONTAL);
			Mat1f SAT_P;
			if(flow_parameters->filter_method == TCParameters::kMethod_NC || flow_parameters->filter_method == TCParameters::kMethod_IC)
			{
				SAT_P = cumsum<float>(dPdt, kHORIZONTAL);
			}

			for(int a = 0; a < additional_information.size(); ++a)
			{
			
				// Generate the domain transform filtering input
				Mat1f path_additional_information(1, (int)active_path_size);
				for(int p = 0; p < position.size(); ++p)
				{
					path_additional_information.at<float>(0, p) = *additional_information[a][p];
				}

				// Apply the right filter
				Mat1f filtered_path_additional_information;
				if(flow_parameters->filter_method == TCParameters::kMethod_RF)
				{
					filtered_path_additional_information =RecursiveFiltering1d<float>(path_additional_information, dPdt, additional_information_parameters->sigma_st, additional_information_parameters->iterations);
				}
				else if(flow_parameters->filter_method == TCParameters::kMethod_NC)
				{
					filtered_path_additional_information = NormalizedConvolution1d<float>(path_additional_information, SAT_P, additional_information_parameters->sigma_st, additional_information_parameters->iterations);
				}
				else
				{
					filtered_path_additional_information = InterpolatedConvolution1d<float>(path_additional_information, SAT_P, additional_information_parameters->sigma_st, additional_information_parameters->iterations);
				}

				// Update the values
				for(int p = 0; p < active_path_size; ++p)
				{
					*additional_information[a][p] = filtered_path_additional_information.at<float>(0, p);
				}
			}
		}
	}
}
