// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.


////////////////////////////////////////////////////////////////////////////////
//  This file contains an implementation of the temporal consistent filtering 
//  technique. Please see the original paper for details:
//
////    Practical temporal consistency for image-based graphics applications
////    Manuel Lang, Oliver Wang, Tunc Aydin, Aljoscha Smolic, Markus Gross
////    ACM Trans. Graph. 31.4 (2012): 34
////
////  For more information about the paper visit:
////  
////           http://zurich.disneyresearch.com/~malang/projects/FeatureFlow/
////
//
//  This C++ implementation is based on the description contained in the paper;
//  even when it provides good results, we cannot asure it is equivalent to the
//  implementation generated by the original authors, since they did not publish 
//  sample code.
////////////////////////////////////////////////////////////////////////////////

#include <DomainTransform.h>
#include <FlowUtils.h>

#include "PathKernel.h"
#include "TemporalConsistency.h"

using namespace cv;
using namespace std;


// Filters the video volume V[view][frame] in time and between views.
//
// Inspired by the Domain transform <see "DomainTransform/domaintransform.h">.
//
// the filtering follows an iterative approach: 
// First, the solution is initializated with the domain transform from the input images(since it is invariable) 
// and rough estimations of the optical flow and disparity (if there are more than 1 view).
// In each iteration, the optical flow and disparity are estimated using either the output of the previous iteration,
// then, one filter pass is realized in each dimension:
//   1. between views, following the disparity maps 
//   2. in time, following the paths generated by the optical flow
//   3. in space, using the XY filtering provided by the Domain transform
// The filtering generates new approximations of the information contained in the video volume, including:
//   * RGB channels
//   * Optical Flow
//   * Disparity map 
//   * Additional information
// Which in turn will be used as input for the next iteration.
vector< vector<KeyFrame> > TemporalConsistency(const vector< vector<KeyFrame> >& input, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters)
{
	vector< vector<KeyFrameTC> > keyframes = importKeyframes(input);

	// Transform domain
	calculateTransformDomain(keyframes, flow_parameters, disparity_parameters, additional_information_parameters);

	// Compute confidence G0
	initializeConfidence(keyframes, flow_parameters.filter, disparity_parameters.filter);

	// Initial Spatial Filtering
	filterXY(keyframes, flow_parameters, disparity_parameters, additional_information_parameters);

	int iterations = max(max(flow_parameters.iterations, disparity_parameters.iterations), additional_information_parameters.iterations);

	for(int iteration = 0; iteration < iterations; ++iteration)
	{
		// Dense flow estimate 
		estimateOpticalFlow(keyframes, flow_parameters.filter, disparity_parameters.filter);

		// Oclussion Estimates
		applyPenalityFlow(keyframes, flow_parameters, disparity_parameters, additional_information_parameters, iteration);

		// Multi view filtering
		TCParameters no_parameters = TCParameters::off();
		filterV(keyframes, no_parameters, disparity_parameters, additional_information_parameters);

		// Temporal Filtering
		filterT(keyframes, flow_parameters, disparity_parameters, additional_information_parameters);

		// Spatial Filtering
		filterXY(keyframes, flow_parameters, disparity_parameters, additional_information_parameters);
	}

	// Final dense flow estimate 
	estimateOpticalFlow(keyframes, flow_parameters.filter, disparity_parameters.filter);
	
	return exportKeyframes(keyframes);
}

// Computes the domain transform of each frame for each filtering method
inline void calculateTransformDomain(vector< vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters)
{
	#pragma omp parallel for
	for (int f = 0; f < keyframes[0].size(); ++f)
	{
		for (int v = 0; v < keyframes.size(); ++v)
		{
			if(flow_parameters.filter)
			{
				float luminance_scale = 1.0f;
				if(!flow_parameters.luminance_scale.empty())
					luminance_scale = flow_parameters.luminance_scale[v];

				Mat1f dHdx = DomainTransform<Vec3f>(*keyframes[v][f].current, flow_parameters.sigma_r * luminance_scale, flow_parameters.sigma_s, kHORIZONTAL);
				Mat1f dVdy = DomainTransform<Vec3f>(*keyframes[v][f].current, flow_parameters.sigma_r * luminance_scale, flow_parameters.sigma_s, kVERTICAL);

				// saves the domain transform or the SAT acording to the filtering method
				if(flow_parameters.filter_method == TCParameters::kMethod_RF)
				{
					keyframes[v][f].flow_ct_H = dHdx;
					keyframes[v][f].flow_ct_V = dVdy;
				}
				else
				{
					keyframes[v][f].flow_ct_H = cumsum<float>(dHdx, kHORIZONTAL);
					keyframes[v][f].flow_ct_V = cumsum<float>(dVdy, kVERTICAL);
				}
			}

			if(disparity_parameters.filter)
			{
				float luminance_scale = 1.0f;
				if(!disparity_parameters.luminance_scale.empty())
					luminance_scale = disparity_parameters.luminance_scale[v];

				Mat1f dHdx = DomainTransform<Vec3f>(*keyframes[v][f].current, disparity_parameters.sigma_r * luminance_scale, disparity_parameters.sigma_s, kHORIZONTAL);
				Mat1f dVdy = DomainTransform<Vec3f>(*keyframes[v][f].current, disparity_parameters.sigma_r * luminance_scale, disparity_parameters.sigma_s, kVERTICAL);
				
				if(disparity_parameters.filter_method == TCParameters::kMethod_RF)
				{
					keyframes[v][f].disparity_ct_H = dHdx;
					keyframes[v][f].disparity_ct_V = dVdy;
				}
				else
				{
					keyframes[v][f].disparity_ct_H = cumsum<float>(dHdx, kHORIZONTAL);
					keyframes[v][f].disparity_ct_V = cumsum<float>(dVdy, kVERTICAL);
				}
			}

			if(additional_information_parameters.filter)
			{
				float luminance_scale = 1.0f;
				if(!additional_information_parameters.luminance_scale.empty())
					luminance_scale = additional_information_parameters.luminance_scale[v];

				Mat1f dHdx = DomainTransform<Vec3f>(*keyframes[v][f].current, additional_information_parameters.sigma_r * luminance_scale, additional_information_parameters.sigma_s, kHORIZONTAL);
				Mat1f dVdy = DomainTransform<Vec3f>(*keyframes[v][f].current, additional_information_parameters.sigma_r * luminance_scale, additional_information_parameters.sigma_s, kVERTICAL);

				if(additional_information_parameters.filter_method == TCParameters::kMethod_RF)
				{
					keyframes[v][f].additional_ct_H = dHdx;
					keyframes[v][f].additional_ct_V = dVdy;
				}
				else
				{
					keyframes[v][f].additional_ct_H = cumsum<float>(dHdx, kHORIZONTAL);
					keyframes[v][f].additional_ct_V = cumsum<float>(dVdy, kVERTICAL);
				}
			}
		}
	}
}

template <class TValue>
Mat_<TValue> applyConfidence(const Mat_<TValue>& image, const Mat1f& mask)
{
	Mat_<TValue> result(image.rows, image.cols);
	
	result = applyMask<TValue>(image, applyMask<TValue>(getMask<TValue>(image, Mat_<TValue>::zeros(image.rows, image.cols) + kMOVEMENT_UNKNOWN, 0.000001f), mask));
	return result;
}

inline void initializeConfidence(vector< vector<KeyFrameTC> >& keyframes, bool filter_flow, bool filter_disparity)
{
	#pragma omp parallel for
	for (int f = 0; f < keyframes[0].size(); ++f)
	{
		for (int v = 0; v < keyframes.size(); ++v)
		{
			if(filter_flow)
			{
				size_t n = keyframes[v].size();

				if(f+1 < n)
				{
					keyframes[v][f].flow_confidence_forward = getFlowConfidence(keyframes[v][f].flow_forward, keyframes[v][f + 1].flow_backward);
					keyframes[v][f].flow_forward = applyConfidence<Vec2f>(keyframes[v][f].flow_forward, keyframes[v][f].flow_confidence_forward);
				}
				else
				{
					// keyframes[v][f].flow_forward = Mat2f::zeros(keyframes[v][f].current->rows, keyframes[v][f].current->cols);
					keyframes[v][f].flow_confidence_forward = Mat1f::zeros(keyframes[v][f].current->rows, keyframes[v][f].current->cols);
				}

				if(f > 0)
				{
					keyframes[v][f].flow_confidence_backward = getFlowConfidence(keyframes[v][f].flow_backward, keyframes[v][f - 1].flow_forward);
					keyframes[v][f].flow_backward = applyConfidence<Vec2f>(keyframes[v][f].flow_backward, keyframes[v][f].flow_confidence_backward);
				}
				else
				{
					// keyframes[v][f].flow_backward = Mat2f::zeros(keyframes[v][f].current->rows, keyframes[v][f].current->cols);
					keyframes[v][f].flow_confidence_backward = Mat1f::zeros(keyframes[v][f].current->rows, keyframes[v][f].current->cols);
				}
			}

			if(filter_disparity)
			{
				size_t n = keyframes.size();

				if(v+1 < n)
				{
					keyframes[v][f].disparity_confidence_forward = getDisparityConfidenceBinary(keyframes[v][f].disparity_forward, keyframes[v + 1][f].disparity_backward);
					keyframes[v][f].disparity_forward = applyConfidence<float>(keyframes[v][f].disparity_forward, keyframes[v][f].disparity_confidence_forward);
				}

				if(v > 0)
				{
					keyframes[v][f].disparity_confidence_backward = getDisparityConfidenceBinary(keyframes[v][f].disparity_backward, keyframes[v - 1][f].disparity_forward);
					keyframes[v][f].disparity_backward = applyConfidence<float>(keyframes[v][f].disparity_backward, keyframes[v][f].disparity_confidence_backward);
				}
			}
		}

		size_t nviews = keyframes.size();
		keyframes[0][f].disparity_backward            = -keyframes[0][f].disparity_forward;
		keyframes[0][f].disparity_confidence_backward =  keyframes[0][f].disparity_confidence_forward;
		keyframes[nviews - 1][f].disparity_forward            = -keyframes[nviews - 1][f].disparity_backward;
		keyframes[nviews - 1][f].disparity_confidence_forward =  keyframes[nviews - 1][f].disparity_confidence_backward;
	}
}

inline Mat2f normalizeFlow(Mat2f flow, Mat1f confidence)
{
	// Known flow mask.
	Mat1f mask = getMask<Vec2f>(flow, Mat2f::zeros(flow.rows, flow.cols) + kMOVEMENT_UNKNOWN, 0.000001f, 1);
	Mat2f flow_masked = applyMask<Vec2f>(flow, mask);

	if(!confidence.empty())
	{
		mask = applyMask<float>(confidence, mask);
	}

	Mat2f flow_propagated = normalize<Vec2f>(flow_masked, mask);
	return flow_propagated;
}

inline Mat2f normalizeDisparity(Mat1f disparity, Mat1f confidence)
{
	// Known disparity mask.
	Mat1f mask = getMask<float>(disparity, Mat1f::zeros(disparity.rows, disparity.cols) + kMOVEMENT_UNKNOWN, 0.000001f, 1);
	Mat1f disparity_masked = applyMask<float>(disparity, mask);

	if(!confidence.empty())
	{
		mask = applyMask<float>(confidence, mask);
	}

	Mat1f disparity_propagated = normalize<float>(disparity_masked, mask);
	return disparity_propagated;
}

inline void estimateOpticalFlow(vector< vector<KeyFrameTC> >& keyframes, bool flow, bool disparity)
{
	size_t nframes = keyframes[0].size();
	size_t nviews = keyframes.size();

	#pragma omp parallel for
	for (int f = 0; f < nframes; ++f)
	{
		keyframes[0][f].disparity_backward            = -keyframes[0][f].disparity_forward;
		keyframes[0][f].disparity_confidence_backward =  keyframes[0][f].disparity_confidence_forward;
		keyframes[nviews-1][f].disparity_forward            = -keyframes[nviews - 1][f].disparity_backward;
		keyframes[nviews-1][f].disparity_confidence_forward =  keyframes[nviews - 1][f].disparity_confidence_backward;

		for (int v = 0; v < nviews; ++v)
		{
			if(flow)
			{
				keyframes[v][f].flow_estimation_forward  = normalizeFlow(keyframes[v][f].flow_forward,  keyframes[v][f].flow_confidence_forward);
				keyframes[v][f].flow_estimation_backward = normalizeFlow(keyframes[v][f].flow_backward, keyframes[v][f].flow_confidence_backward);
			}

			if(disparity)
			{
				keyframes[v][f].disparity_estimation_forward  = normalizeDisparity(keyframes[v][f].disparity_forward,  keyframes[v][f].disparity_confidence_forward);
				keyframes[v][f].disparity_estimation_backward = normalizeDisparity(keyframes[v][f].disparity_backward, keyframes[v][f].disparity_confidence_backward);
			}
		}
	}
}

inline void applyPenalityFlow(vector< vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters, int iteration)
{
	#pragma omp parallel for
	for (int f = 0; f < keyframes[0].size(); ++f)
	{
		for (int v = 0; v < keyframes.size(); ++v)
		{
			// Occlusion estimates
			KeyFrameTC& keyframe = keyframes[v][f];
			if(flow_parameters.filter)
			{
				if(f + 1 < keyframes[0].size())
				{
					Mat1f flow_penality_forward = getFlowConfidence(keyframe.flow_estimation_forward, keyframes[v][f + 1].flow_estimation_backward);
					pow(flow_penality_forward, flow_parameters.theta, flow_penality_forward);

					keyframe.flow_forward = applyMask<Vec2f>(keyframe.flow_forward, flow_penality_forward);
					keyframe.flow_confidence_forward = keyframe.flow_confidence_forward.mul(flow_penality_forward);
					keyframe.flow_penality_forward = flow_penality_forward;
				}

				if(f > 0)
				{
					Mat1f flow_penality_backward = getFlowConfidence(keyframe.flow_estimation_backward, keyframes[v][f - 1].flow_estimation_forward);
					pow(flow_penality_backward, flow_parameters.theta, flow_penality_backward);

					keyframe.flow_backward = applyMask<Vec2f>(keyframe.flow_backward, flow_penality_backward);
					keyframe.flow_confidence_backward = keyframe.flow_confidence_backward.mul(flow_penality_backward);
					keyframe.flow_penality_backward = flow_penality_backward;
				}
			}
			
			if(disparity_parameters.filter)
			{
				float theta = max(0.0f, disparity_parameters.theta - iteration);
				if(v + 1 < keyframes.size())
				{	
					Mat1f disparity_penality_forward = getDisparityConfidenceBinary(keyframe.disparity_estimation_forward, keyframes[v + 1][f].disparity_estimation_backward, 4, 0.25f);
					pow(disparity_penality_forward, theta, disparity_penality_forward);
					
					keyframe.disparity_forward            = applyMask<float>(keyframe.disparity_forward,            disparity_penality_forward);
					keyframe.disparity_confidence_forward = applyMask<float>(keyframe.disparity_confidence_forward, disparity_penality_forward);
					keyframe.disparity_penality_forward   = disparity_penality_forward;
				}

				if(v > 0)
				{
					Mat1f disparity_penality_backward = getDisparityConfidenceBinary(keyframe.disparity_estimation_backward, keyframes[v - 1][f].disparity_estimation_forward, 4, 0.25f);
					pow(disparity_penality_backward, theta, disparity_penality_backward);

					keyframe.disparity_backward            = applyMask<float>(keyframe.disparity_backward,            disparity_penality_backward);
					keyframe.disparity_confidence_backward = applyMask<float>(keyframe.disparity_confidence_backward, disparity_penality_backward);
					keyframe.disparity_penality_backward   = disparity_penality_backward;
				}
			}
		}
	}
}

inline void filterXY(vector< vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters)
{
	#pragma omp parallel for
	for (int f = 0; f < keyframes[0].size(); ++f)
	{
		for (int v = 0; v < keyframes.size(); ++v)
		{
			if(flow_parameters.filter)
			{
				if(flow_parameters.filter_method == TCParameters::kMethod_RF)
				{
					keyframes[v][f].flow_forward             = RecursiveFiltering2d<Vec2f>(keyframes[v][f].flow_forward,             keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
					keyframes[v][f].flow_confidence_forward  = RecursiveFiltering2d<float>(keyframes[v][f].flow_confidence_forward,  keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);

					keyframes[v][f].flow_backward            = RecursiveFiltering2d<Vec2f>(keyframes[v][f].flow_backward,            keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
					keyframes[v][f].flow_confidence_backward = RecursiveFiltering2d<float>(keyframes[v][f].flow_confidence_backward, keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
				}
				else if(flow_parameters.filter_method == TCParameters::kMethod_NC)
				{
					keyframes[v][f].flow_forward             = NormalizedConvolution2d<Vec2f>(keyframes[v][f].flow_forward,             keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
					keyframes[v][f].flow_confidence_forward  = NormalizedConvolution2d<float>(keyframes[v][f].flow_confidence_forward,  keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);

					keyframes[v][f].flow_backward            = NormalizedConvolution2d<Vec2f>(keyframes[v][f].flow_backward,            keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
					keyframes[v][f].flow_confidence_backward = NormalizedConvolution2d<float>(keyframes[v][f].flow_confidence_backward, keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
				}
				else // TCParameters::kMethod_IC
				{
					keyframes[v][f].flow_forward             = InterpolatedConvolution2d<Vec2f>(keyframes[v][f].flow_forward,             keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
					keyframes[v][f].flow_confidence_forward  = InterpolatedConvolution2d<float>(keyframes[v][f].flow_confidence_forward,  keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);

					keyframes[v][f].flow_backward            = InterpolatedConvolution2d<Vec2f>(keyframes[v][f].flow_backward,            keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
					keyframes[v][f].flow_confidence_backward = InterpolatedConvolution2d<float>(keyframes[v][f].flow_confidence_backward, keyframes[v][f].flow_ct_H, keyframes[v][f].flow_ct_V, flow_parameters.sigma_s, flow_parameters.propagation_iterations);
				}
			}

			if(disparity_parameters.filter)
			{
				if(disparity_parameters.filter_method == TCParameters::kMethod_RF)
				{
					keyframes[v][f].disparity_forward             = RecursiveFiltering2d<float>(keyframes[v][f].disparity_forward,             keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
					keyframes[v][f].disparity_confidence_forward  = RecursiveFiltering2d<float>(keyframes[v][f].disparity_confidence_forward,  keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
		
					keyframes[v][f].disparity_backward            = RecursiveFiltering2d<float>(keyframes[v][f].disparity_backward,            keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
					keyframes[v][f].disparity_confidence_backward = RecursiveFiltering2d<float>(keyframes[v][f].disparity_confidence_backward, keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
				}
				else if(disparity_parameters.filter_method == TCParameters::kMethod_NC)
				{
					keyframes[v][f].disparity_forward             = NormalizedConvolution2d<float>(keyframes[v][f].disparity_forward,             keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
					keyframes[v][f].disparity_confidence_forward  = NormalizedConvolution2d<float>(keyframes[v][f].disparity_confidence_forward,  keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
		
					keyframes[v][f].disparity_backward            = NormalizedConvolution2d<float>(keyframes[v][f].disparity_backward,            keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
					keyframes[v][f].disparity_confidence_backward = NormalizedConvolution2d<float>(keyframes[v][f].disparity_confidence_backward, keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
				}
				else // TCParameters::kMethod_IC
				{
					keyframes[v][f].disparity_forward             = InterpolatedConvolution2d<float>(keyframes[v][f].disparity_forward,             keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
					keyframes[v][f].disparity_confidence_forward  = InterpolatedConvolution2d<float>(keyframes[v][f].disparity_confidence_forward,  keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
		
					keyframes[v][f].disparity_backward            = InterpolatedConvolution2d<float>(keyframes[v][f].disparity_backward,            keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
					keyframes[v][f].disparity_confidence_backward = InterpolatedConvolution2d<float>(keyframes[v][f].disparity_confidence_backward, keyframes[v][f].disparity_ct_H, keyframes[v][f].disparity_ct_V, disparity_parameters.sigma_s, disparity_parameters.propagation_iterations);
				}
			}

			if(additional_information_parameters.filter && !keyframes[v][f].additional_information.empty())
			{
				for(int a = 0; a < keyframes[v][f].additional_information.size(); ++a)
				{
					if(additional_information_parameters.filter_method == TCParameters::kMethod_RF)
					{
						keyframes[v][f].additional_information[a] = RecursiveFiltering2d<float>(keyframes[v][f].additional_information[a], keyframes[v][f].additional_ct_H, keyframes[v][f].additional_ct_V, additional_information_parameters.sigma_s, additional_information_parameters.propagation_iterations);
					}
					else if(additional_information_parameters.filter_method == TCParameters::kMethod_NC)
					{
						keyframes[v][f].additional_information[a] = NormalizedConvolution2d<float>(keyframes[v][f].additional_information[a], keyframes[v][f].additional_ct_H, keyframes[v][f].additional_ct_V, additional_information_parameters.sigma_s, additional_information_parameters.propagation_iterations);
					}
					else //  TCParameters::kMethod_IC
					{
						keyframes[v][f].additional_information[a] = InterpolatedConvolution2d<float>(keyframes[v][f].additional_information[a], keyframes[v][f].additional_ct_H, keyframes[v][f].additional_ct_V, additional_information_parameters.sigma_s, additional_information_parameters.propagation_iterations);
					}
				}
			}
		}
	}
}

inline vector< vector<KeyFrameTC> > importKeyframes(const vector< vector<KeyFrame> >& input)
{
	vector< vector<KeyFrameTC> > keyframes(input.size());

	// Reserve the memory.
	for (int v = 0; v < input.size(); ++v)
	{
		keyframes[v] = vector<KeyFrameTC>(input[v].size());
	}

	// Copy the information.
	for (int f = 0; f < input[0].size(); ++f)
	{
		for (int v = 0; v < input.size(); ++v)
		{
			keyframes[v][f].number                        = input[v][f].number;
			keyframes[v][f].current                       = input[v][f].current;
			keyframes[v][f].next                          = input[v][f].next;
			keyframes[v][f].previous                      = input[v][f].previous;
			keyframes[v][f].flow_forward                  = input[v][f].flow_forward;
			keyframes[v][f].flow_backward                 = input[v][f].flow_backward;
			keyframes[v][f].flow_estimation_forward       = input[v][f].flow_forward;
			keyframes[v][f].flow_estimation_backward      = input[v][f].flow_backward;
			keyframes[v][f].disparity_forward             = input[v][f].disparity_forward;
			keyframes[v][f].disparity_backward            = input[v][f].disparity_backward;
			keyframes[v][f].disparity_estimation_forward  = input[v][f].disparity_forward;
			keyframes[v][f].disparity_estimation_backward = input[v][f].disparity_backward;
			keyframes[v][f].additional_information        = input[v][f].additional_information;
		}
	}

	return keyframes;
}

inline vector< vector<KeyFrame> > exportKeyframes(const vector< vector<KeyFrameTC> >& keyframes)
{
	vector< vector<KeyFrame> > result(keyframes.size());
	for (int v = 0; v < keyframes.size(); ++v)
	{
		result[v] = vector<KeyFrame>(keyframes[v].size());
	}

	for (int f = 0; f < keyframes[0].size(); ++f)
	{
		for (int v = 0; v < keyframes.size(); ++v)
		{
			result[v][f].number                        = keyframes[v][f].number;
			result[v][f].current                       = keyframes[v][f].current;
			result[v][f].next                          = keyframes[v][f].next;
			result[v][f].previous                      = keyframes[v][f].previous;
			result[v][f].flow_forward                  = keyframes[v][f].flow_estimation_forward;
			result[v][f].flow_backward                 = keyframes[v][f].flow_estimation_backward;
			result[v][f].disparity_forward             = keyframes[v][f].disparity_estimation_forward;
			result[v][f].disparity_backward            = keyframes[v][f].disparity_estimation_backward;
			result[v][f].disparity_confidence_forward  = keyframes[v][f].disparity_confidence_forward;
			result[v][f].disparity_confidence_backward = keyframes[v][f].disparity_confidence_backward;
			result[v][f].additional_information        = keyframes[v][f].additional_information;
		}
	}

	return result;
}

inline void filterT(vector< vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters)
{
	// box dimensions
	int h = keyframes[0][0].current->rows;
	int w = keyframes[0][0].current->cols;
	size_t nframes = keyframes[0].size();
	int iterations = 1;

	for(int iteration = 0; iteration < iterations; ++iteration)
	{
		// filtering parameters
		if(flow_parameters.filter)
			flow_parameters.sigma_h = SigmaHi(flow_parameters.sigma_st, iteration, flow_parameters.iterations);
		if(disparity_parameters.filter)
			disparity_parameters.sigma_h = SigmaHi(disparity_parameters.sigma_st, iteration, disparity_parameters.iterations);
		if(additional_information_parameters.filter)
			additional_information_parameters.sigma_h = SigmaHi(additional_information_parameters.sigma_st, iteration, additional_information_parameters.iterations);
		
		for(int v= 0; v < keyframes.size(); ++v)
		{
			// Initialize the paths.
			vector<Mat_<PathKernel*> > paths(3);

			{// Forward Time filtering
				// populate the first frame with new Kernels
				paths[0] = Mat_<PathKernel*>(h, w);
				#pragma omp parallel for
				for(int y = 0; y < h; ++y)
				{
					for(int x = 0; x < w; ++x)
					{
						PathKernel* kernel = new PathKernel(&flow_parameters, &disparity_parameters, &additional_information_parameters);
						paths[0].at<PathKernel*>(y, x) = kernel;
						kernel->updatePath(keyframes[v][0], y, x);
					}
				}

				int path_count = h * w;
				int total_path_count = path_count;

				// populate the the second frame (next) and third frame (cleaning) with NULL
				for(int f = 1; f <= 2; ++f)
				{
					// TODO: try another constructor like Mat_<PathKernel*>(h, w, NULL)
					paths[f] = Mat_<PathKernel*>(h, w);
					for(int y = 0; y < h; ++y)
					{
						for(int x = 0; x < w; ++x)
						{
							paths[f].at<PathKernel*>(y, x) = NULL;
						}
					}
				}

				// Forward filtering.
				for(int f = 0; f < nframes; ++f)
				{
					#pragma omp parallel for
					for(int y = 0; y < h; ++y)
					{
						for(int x = 0; x < w; ++x)
						{
							PathKernel * kernel = NULL;
							bool new_path = false;
							bool end_of_path = false;

							// new path
							if (f > 0 && paths[0].at<PathKernel*>(y, x) == NULL)
							{
								// new path: double check
								// TODO optimize replacing with atomic check and swap
								#pragma omp critical
								{
									if(paths[0].at<PathKernel*>(y, x) == NULL)
									{
										// Associate the pixel with a new kernel
										kernel = new PathKernel(&flow_parameters, &disparity_parameters, &additional_information_parameters);
										paths[0].at<PathKernel*>(y, x) = kernel;
										new_path = true;
									}
								}
								
								if(new_path)
								//if(kernel != NULL)
								{
									// Complete the fields outside the critical section
									kernel->updatePath(keyframes[v][f], y, x);
									#pragma omp atomic
									path_count++;
									#pragma omp atomic
									total_path_count++;
								}
							}

							if(kernel == NULL)
							{
								kernel = paths[0].at<PathKernel*>(y, x);
							}

							end_of_path = true; // it's the end of the path, unless there is a valid position on the next frame
							if(f + 1 < nframes)
							{
								Vec2f flow_vector = keyframes[v][f].flow_estimation_forward.at<Vec2f>(y, x);
								Vec2i position_next(cvRound(y + flow_vector[1]), cvRound(x + flow_vector[0]));

								// Check if the path leads outside the screen.
								if(position_next[0] >= 0 && position_next[0] < h && position_next[1] >= 0 && position_next[1] < w)
								{
									// check for collisions
									if (paths[1].at<PathKernel*>(position_next[0], position_next[1]) == NULL)
									{
										// check for collisions: double check
										// TODO optimize replacing with atomic check and swap
										#pragma omp critical
										{
											if(paths[1].at<PathKernel*>(position_next[0], position_next[1]) == NULL)
											{
												// Asociate the next pixel with the kernel.
												paths[1].at<PathKernel*>(position_next[0], position_next[1]) = kernel;
												end_of_path = false;
											}
										}
										if(!end_of_path)
										{
											// Complete the fields outside the critical section.
											kernel->updatePath(keyframes[v][f + 1], position_next[0], position_next[1]);
										}
									}
								}
							}

							if(end_of_path)
							{
								// filter the path
								kernel->apply(PathKernel::kFilterTime, v);
					
								// Remove the kernel from the memory
								delete kernel;
								#pragma omp atomic
								path_count--;
							}
						}
					}

					paths[0] = paths[1];
					paths[1] = paths[2].clone();
				}
			}

			{// Backward Time filtering
				// populate the first frame with new Kernels
				paths[0] = Mat_<PathKernel*>(h, w);
				#pragma omp parallel for
				for(int y = 0; y < h; ++y)
				{
					for(int x = 0; x < w; ++x)
					{
						size_t f = nframes-1;
						PathKernel * kernel = new PathKernel(&flow_parameters, &disparity_parameters, &additional_information_parameters);
						paths[0].at<PathKernel*>(y, x) = kernel;
						kernel->updatePath(keyframes[v][f],y, x);
					}
				}

				int path_count = h * w;
				int total_path_count = path_count;

				// populate the the second frame (next) and third frame (cleaning) with NULL
				for(int f = 1; f <= 2; ++f)
				{
					paths[f] = Mat_<PathKernel*>(h, w);
					for(int y = 0; y < h; ++y)
					{
						for(int x = 0; x < w; ++x)
						{
							paths[f].at<PathKernel*>(y, x) = NULL;
						}
					}
				}

				// Backward filtering
				for(int f = (int)keyframes.size() - 1; f >= 0 ; --f)
				{
					#pragma omp parallel for
					for(int y = 0; y < h; ++y)
					{
						for(int x = 0; x < w; ++x)
						{		
							PathKernel * kernel = NULL;
							bool new_path = false;
							bool end_of_path = false;

							// new path
							if (f < keyframes[v].size() - 1 && paths[0].at<PathKernel*>(y, x) == NULL)
							{
								// new path: double check
								// TODO optimize replacing with atomic check and swap
								#pragma omp critical
								{
									if(paths[0].at<PathKernel*>(y, x) == NULL)
									{
										// Associate the pixel with a new kernel
										kernel = new PathKernel(&flow_parameters, &disparity_parameters, &additional_information_parameters);
										paths[0].at<PathKernel*>(y, x) = kernel;
										new_path = true;
									}
								}
								if(new_path)
								{
									// Complete the fields outside the critical section
									kernel->updatePath(keyframes[v][f], y, x);
									#pragma omp atomic
									path_count++;
									#pragma omp atomic
									total_path_count++;
								}
							}

							if(!new_path)
							{
								kernel = paths[0].at<PathKernel*>(y, x);
							}

							end_of_path = true; // it's the end of the path, unless there is a valid position on the previous frame

							if(f - 1 >= 0)
							{
								Vec2f flow_vector = keyframes[v][f].flow_estimation_backward.at<Vec2f>(y, x);
								Vec2i position_previous(cvRound(y + flow_vector[1]), cvRound(x + flow_vector[0]));

								// Check if the path leads outside the screen.
								if(position_previous[0] >= 0 && position_previous[0] < h && position_previous[1] >= 0 && position_previous[1] < w)
								{
									// check for collisions
									if (paths[1].at<PathKernel*>(position_previous[0], position_previous[1]) == NULL)
									{
										// check for collisions: double check
										// TODO optimize replacing with atomic check and swap
										#pragma omp critical
										{
											if(paths[1].at<PathKernel*>(position_previous[0], position_previous[1]) == NULL)
											{
												// Asociate the next pixel with the kernel
												paths[1].at<PathKernel*>(position_previous[0], position_previous[1]) = kernel;
												end_of_path = false;
											}
										}

										if(!end_of_path)
										{
											// Complete the fields outside the critical section
											kernel->updatePath(keyframes[v][f-1], position_previous[0], position_previous[1]);
										}
									}
								}
							}

							if(end_of_path)
							{
								// filter the path
								kernel->apply(PathKernel::kFilterTime, v);
								// Remove the kernel from the memory
								delete kernel;
								#pragma omp atomic
								path_count--;
							}
						}
					}

					paths[0] = paths[1];
					paths[1] = paths[2].clone();
				}
			}
		}
	}
}

inline void filterV(vector< vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters)
{
	if(keyframes.size() < 2 || !disparity_parameters.filter)
		return;

	// box dimensions
	int h = keyframes[0][0].current->rows;
	int w = keyframes[0][0].current->cols;
	size_t nviews = keyframes.size();
	size_t nframes = keyframes[0].size();
	int iterations = 1;

	for(int iteration = 0; iteration < iterations; ++iteration)
	{
		// filtering parameters
		if(flow_parameters.filter)
			flow_parameters.sigma_h =  SigmaHi(flow_parameters.sigma_st, iteration, flow_parameters.iterations);

		if(disparity_parameters.filter)
			disparity_parameters.sigma_h =  SigmaHi(disparity_parameters.sigma_st, iteration, disparity_parameters.iterations);

		if(additional_information_parameters.filter)
			additional_information_parameters.sigma_h =  SigmaHi(additional_information_parameters.sigma_st, iteration, additional_information_parameters.iterations);
		
		for(int f= 0; f < nframes; ++f)
		{
			// Initialize the paths
			vector<Mat_<PathKernel*> > paths(3);

			{// Forward Time filtering
				// populate the first frame with new Kernels
				paths[0] = Mat_<PathKernel*>(h, w);
				#pragma omp parallel for
				for(int y = 0; y < h; ++y)
				{
					for(int x = 0; x < w; ++x)
					{
						int v = 0;
						PathKernel * kernel = new PathKernel(&flow_parameters, &disparity_parameters, &additional_information_parameters);
						paths[0].at<PathKernel*>(y, x) = kernel;
						kernel->updatePath(keyframes[v][f], y, x);
					}
				}

				int path_count = h * w;
				int total_path_count = path_count;

				// populate the the second view (next) and third view (cleaning) with NULL
				for(int v = 1; v <= 2; ++v)
				{
					paths[v] = Mat_<PathKernel*>(h, w);
					#pragma omp parallel for
					for(int y = 0; y < h; ++y)
					{
						for(int x = 0; x < w; ++x)
						{
							paths[v].at<PathKernel*>(y, x) = NULL;
						}
					}
				}

				// Forward filtering
				for(int v = 0; v < nviews; ++v)
				{
					#pragma omp parallel for
					for(int y = 0; y < h; ++y)
					{
						for(int x = 0; x < w; ++x)
						{
							PathKernel * kernel = NULL;
							bool new_path = false;
							bool end_of_path = false;

							// new path
							if (v > 0 && paths[0].at<PathKernel*>(y, x) == NULL)
							{
								// new path: double check
								// TODO optimize replacing with atomic check and swap
								#pragma omp critical
								{
									if(paths[0].at<PathKernel*>(y, x) == NULL)
									{	// Associate the pixel with a new kernel
										kernel = new PathKernel(&flow_parameters, &disparity_parameters, &additional_information_parameters);
										paths[0].at<PathKernel*>(y, x) = kernel;
										new_path = true;
									}
								}
								if(new_path)
								{	// Complete the fields outside the critical section
									kernel->updatePath(keyframes[v][f], y, x);
									#pragma omp atomic
									path_count++;
									#pragma omp atomic
									total_path_count++;
								}
							}

							if(!new_path)
							{
								kernel = paths[0].at<PathKernel*>(y, x);
							}

							end_of_path = true; // it's the end of the path, unless there is a valid position on the next frame

							if(v + 1 < nviews)
							{
								Vec2f flow_vector = Vec2f(keyframes[v][f].disparity_estimation_forward.at<float>(y, x), 0);
								Vec2i position_next(cvRound(y + flow_vector[1]), cvRound(x + flow_vector[0]));

								// Check if the path leads outside the screen.
								if(position_next[0] >= 0 && position_next[0] < h && position_next[1] >= 0 && position_next[1] < w)
								{
									// check for collisions
									if (paths[1].at<PathKernel*>(position_next[0], position_next[1]) == NULL)
									{
										// check for collisions: double check
										// TODO optimize replacing with atomic check and swap
										#pragma omp critical
										{
											if(paths[1].at<PathKernel*>(position_next[0], position_next[1]) == NULL)
											{
												// Asociate the next pixel with the kernel.
												paths[1].at<PathKernel*>(position_next[0], position_next[1]) = kernel;
												end_of_path = false;
											}
										}

										if(!end_of_path)
										{
											// Complete the fields outside the critical section.
											kernel->updatePath(keyframes[v+1][f], position_next[0], position_next[1]);
										}
									}
								}
							}

							if(end_of_path)
							{
								// Filter the path.
								kernel->apply(PathKernel::kFilterViews, v);

								// Remove the kernel from the memory.
								delete kernel;
								#pragma omp atomic
								path_count--;
							}
						}
					}

					paths[0] = paths[1];
					paths[1] = paths[2].clone();
				}
			}

			{// Backward Time filtering
				// populate the first frame with new Kernels
				paths[0] = Mat_<PathKernel*>(h, w);	
				#pragma omp parallel for
				for(int y = 0; y < h; ++y)
				{
					for(int x = 0; x < w; ++x)
					{
						size_t v = nviews - 1;
						PathKernel * kernel = new PathKernel(&flow_parameters, &disparity_parameters, &additional_information_parameters);
						paths[0].at<PathKernel*>(y, x) = kernel;
						kernel->updatePath(keyframes[v][f], y, x);
					}
				}

				int path_count = h * w;
				int total_path_count = path_count;

				// populate the the second frame (next) and third frame (cleaning) with NULL
				for(int v = 1; v <= 2; ++v)
				{
					paths[v] = Mat_<PathKernel*>(h, w);
					#pragma omp parallel for
					for(int y = 0; y < h; ++y)
					{
						for(int x = 0; x < w; ++x)
						{
							paths[v].at<PathKernel*>(y, x) = NULL;
						}
					}
				}

				// Backward filtering
				for(int v = (int)nviews-1; v >= 0 ; --v)
				{
					#pragma omp parallel for
					for(int y = 0; y < h; ++y)
					{
						for(int x = 0; x < w; ++x)
						{
							PathKernel * kernel = NULL;
							bool new_path = false;
							bool end_of_path = false;

							// new path
							if (v < nviews - 1 && paths[0].at<PathKernel*>(y, x) == NULL)
							{
								// new path: double check
								// TODO optimize replacing with atomic check and swap
								#pragma omp critical
								{
									if(paths[0].at<PathKernel*>(y, x) == NULL)
									{
										// Associate the pixel with a new kernel.
										kernel = new PathKernel(&flow_parameters, &disparity_parameters, &additional_information_parameters);
										paths[0].at<PathKernel*>(y, x) = kernel;
										new_path = true;
									}
								}

								if(new_path)
								{
									// Complete the fields outside the critical section.
									kernel->updatePath(keyframes[v][f], y, x);
									#pragma omp atomic
									path_count++;
									#pragma omp atomic
									total_path_count++;
								}
							}
							if(!new_path)
							{
								kernel = paths[0].at<PathKernel*>(y, x);
							}

							end_of_path = true; // it's the end of the path, unless there is a valid position on the previous frame

							if(v - 1 >= 0)
							{
								Vec2f flow_vector = Vec2f(keyframes[v][f].disparity_estimation_backward.at<float>(y, x), 0);
								Vec2i position_previous(cvRound(y + flow_vector[1]), cvRound(x + flow_vector[0]));

								// Check if the path leads outside the screen
								if(position_previous[0] >= 0 && position_previous[0] < h && position_previous[1] >= 0 && position_previous[1] < w)
								{
									// check for collisions
									if (paths[1].at<PathKernel*>(position_previous[0], position_previous[1]) == NULL)
									{
										// check for collisions: double check
										// TODO optimize replacing with atomic check and swap
										#pragma omp critical
										{
											if(paths[1].at<PathKernel*>(position_previous[0], position_previous[1]) == NULL)
											{
												// Asociate the next pixel with the kernel
												paths[1].at<PathKernel*>(position_previous[0], position_previous[1]) = kernel;
												end_of_path = false;
											}
										}

										if(!end_of_path)
										{
											// Complete the fields outside the critical section
											kernel->updatePath(keyframes[v-1][f], position_previous[0], position_previous[1]);
										}
									}
								}
							}

							if(end_of_path)
							{
								// filter the path
								kernel->apply(PathKernel::kFilterViews, v);
								// Remove the kernel from the memory
								delete kernel;
								#pragma omp atomic
								path_count--;
							}
						}
					}

					paths[0] = paths[1];
					paths[1] = paths[2].clone();
				}
			}
		}
	}
}

/* DEPRECATED: debug method
void saveInformation(string path, vector< vector<KeyFrameTC> > frames, int iteration)
{
	int    num_digits    = 4;
	float  ERROR_SCALE   = 1.0f;
	float  FLOW_SCALE    = 0.03f;
	float  COMPOSE_SCALE = 0.8f;
	string extension     = ".png";

	if(createDirectory(path.c_str()))
	{
		#pragma omp parallel for
		for(int f = 0; f < frames[0].size(); ++f)
		{
			for(int v = 0; v < frames.size(); ++v)
			{
				// TODO: refactor file name construction.
				{
					ostringstream  ssP;
					ssP << path << "/INPUT" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), float2im(*(frames[v][f].current)));
				}

				if(frames[v][f].next != NULL && !frames[v][f].flow_estimation_forward.empty())
				{
					Mat3f reconstruction =  applyReverseFlow(*(frames[v][f].next),frames[v][f].flow_estimation_forward);
					{
						ostringstream  ssP;
						ssP << path << "/REC" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
						imwrite(ssP.str(), float2im(reconstruction));
					}

					{	
						ostringstream  ssP;
						ssP << path << "/ERROR" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
						imwrite(ssP.str(), float2im(abs(reconstruction-*frames[v][f].current)* ERROR_SCALE));
					}
				}

				{
					ostringstream  ssP;
					ssP << path << "/JFORWARD" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), renderFlow(frames[v][f].flow_forward,frames[v][f].current->clone(),FLOW_SCALE*COMPOSE_SCALE));
				}

				{
					ostringstream  ssP;
					ssP << path << "/JBACKWARD" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), renderFlow(-frames[v][f].flow_backward,frames[v][f].current->clone(),FLOW_SCALE*COMPOSE_SCALE));
				}

				if(!frames[v][f].flow_estimation_forward.empty())
				{
					ostringstream  ssP;
					ssP << path << "/WFORWARD" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), renderFlow(frames[v][f].flow_estimation_forward,frames[v][f].current->clone(),FLOW_SCALE));
				}

				if(!frames[v][f].flow_estimation_backward.empty())
				{
					ostringstream  ssP;
					ssP << path << "/WBACKWARD" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), renderFlow(-frames[v][f].flow_estimation_backward,frames[v][f].current->clone(),FLOW_SCALE));
				}

				Mat1f flow_confidence_forward = frames[v][f].flow_confidence_forward;
				if(!flow_confidence_forward.empty())
				{
					ostringstream  ssP;
					ssP << path << "/CONFF" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), flow_confidence_forward*255*COMPOSE_SCALE);
				}

				Mat1f flow_confidence_backward = frames[v][f].flow_confidence_backward;
				if(!flow_confidence_backward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/CONFB" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), flow_confidence_backward*255*COMPOSE_SCALE);
				}

				Mat1f penality_forward = frames[v][f].flow_penality_forward;
				if(!penality_forward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/PENF" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), penality_forward*255);
				}

				Mat1f penality_backward = frames[v][f].flow_penality_backward;
				if(!penality_backward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/PENB" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), penality_backward*255);
				}

				Mat1f disparity_forward = frames[v][f].disparity_forward;
				if(!disparity_forward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/DF" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), float2im(exportDisparity(disparity_forward)));
				}

				Mat1f disparity_backward = frames[v][f].disparity_backward;
				if(!disparity_backward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/DB" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), float2im(exportDisparity(disparity_backward)));
				}

				Mat1f disparity_confidence_forward = frames[v][f].disparity_confidence_forward;
				if(!disparity_confidence_forward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/DCF" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), (disparity_confidence_forward)*255);
				}

				Mat1f disparity_confidence_backward = frames[v][f].disparity_confidence_backward;
				if(!disparity_confidence_backward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/DCB" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), disparity_confidence_backward*255);
				}

				Mat1f disparity_penality_forward = frames[v][f].disparity_penality_forward;
				if(!disparity_penality_forward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/DPF" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), (disparity_penality_forward)*255);
				}

				Mat1f disparity_penality_backward = frames[v][f].disparity_penality_backward;
				if(!disparity_penality_backward.empty())
				{	
					ostringstream  ssP;
					ssP << path << "/DPB" <<"("<< v<<")"<<std::setw(num_digits) << setfill('0') << f <<"-it"<<iteration<< extension;
					imwrite(ssP.str(), disparity_penality_backward*255);
				}

				if(!frames[v][f].additional_information.empty())
				{
					for(int a = 0; a < frames[v][f].additional_information.size(); ++a)
					{
						ostringstream  ssP;
						ssP << path << "/ADD("<< a <<")" <<std::setw(num_digits) << setfill('0') << f <<"-v"<< v<<"-it"<<iteration<< extension;
						imwrite(ssP.str(), frames[v][f].additional_information[a]*255);
					}
				}
			}
		}
	}
}
*/
