// Copyright Joan Sol Roo and Christian Richardt (7 August 2014)
//
// solroo@gmail.com
// christian@richardt.name
//
// This software is a computer program whose purpose is to compute temporally
// coherent stereoscopic 3D videos from anaglyph 3D input videos.
//
// This software is governed by the CeCILL license under French law and
// abiding by the rules of distribution of free software. You can use, modify
// and/or redistribute the software under the terms of the CeCILL license as
// circulated by CEA, CNRS and INRIA at "http://www.cecill.info".
//
// As a counterpart to the access to the source code and rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty and the software's author, the holder of the
// economic rights, and the successive licensors have only limited liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading, using, modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean that it is complicated to manipulate, and that also
// therefore means that it is reserved for developers and experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and, more generally, to use and operate it in the same
// conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.

#pragma once

#include <opencv2/opencv.hpp>

#include "KeyFrame.h"
#include "KeyFrameTC.h"
#include "TCParameters.h"


// Filters the video volume V[view][frame] in time and between views.
//
// Inspired by the Domain transform <see "DomainTransform/domaintransform.h">.
//
// the filtering follows an iterative approach: 
// First, the solution is initializated with the domain transform from the input images(since it is invariable) 
// and rough estimations of the optical flow and disparity (if there are more than 1 view).
// In each iteration, the optical flow and disparity are estimated using either the output of the previous iteration,
// then, one filter pass is realized in each dimension:
//   1. between views, following the disparity maps 
//   2. in time, following the paths generated by the optical flow
//   3. in space, using the XY filtering provided by the Domain transform
// The filtering generates new approximations of the information contained in the video volume, including:
//   * RGB channels
//   * Optical Flow
//   * Disparity map 
//   * Additional information
// Which in turn will be used as input for the next iteration.
//
// The number of parameters required to configure are too many to explain here, so we suggest you to check the
// DeAnaglyph project, that uses this function.
std::vector< std::vector<KeyFrame> > TemporalConsistency(
	const std::vector< std::vector<KeyFrame> >& input, // Video volume
	TCParameters& flow_parameters, // Flow filtering parameters
	TCParameters& disparity_parameters, // Disparity filtering parameters
	TCParameters& additional_information_parameters // Additional channel fitering parameters
);


//---- Internal functions --------------------------------------------------------------------------

// TODO: replace with normalize(J,G)?
inline cv::Mat2f normalizeFlow(cv::Mat2f flow, cv::Mat1f confidence);
//inline void normalizeConfidence(vector< vector<KeyFrameTC> >& keyframes);

// Computes the domain transform distances for each frame and for each filtering parameter set
inline void calculateTransformDomain(std::vector< std::vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters);

// Computes the initial confidence between forward and backwards flow

inline void initializeConfidence(std::vector< std::vector<KeyFrameTC> >& keyframes, bool filter_flow, bool filter_disparity);

// Estimates the current optical flow based on the last iteration
inline void estimateOpticalFlow(std::vector< std::vector<KeyFrameTC> >& keyframes, bool flow, bool disparity);

// Computes the weighted optical flow, applying a mask based on the current confidence
inline void applyPenalityFlow(std::vector< std::vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters, int iteration);

// Spatial XY pass, using standard Domain transform filtering
inline void filterXY(std::vector< std::vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters);

// Temporal pass, constructing paths (based on the optical flow) and using 1D domain transform filtering
inline void filterT(std::vector< std::vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters);

// Temporal pass, constructing paths (based on the disparity) and using 1D domain transform filtering
inline void filterV(std::vector< std::vector<KeyFrameTC> >& keyframes, TCParameters& flow_parameters, TCParameters& disparity_parameters, TCParameters& additional_information_parameters);

// Transform the input format to Temporal consistency specific, which includes additional data 
inline std::vector< std::vector<KeyFrameTC> > importKeyframes(const std::vector< std::vector<KeyFrame> >& input);

// Transform the interal format into the input format
inline std::vector< std::vector<KeyFrame> > exportKeyframes(const std::vector< std::vector<KeyFrameTC> >& keyframes);

// DEPRECATED: debug method
//void saveInformation(string path, std::vector< std::vector<KeyFrameTC> > frames, int iteration);